"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

import numpy as np
import math, os
import copy, inspect
import pickle
from typing import Iterator

from ANNarchy.core import Global
from ANNarchy.core.Random import RandomDistribution
from ANNarchy.core.Population import Population
from ANNarchy.core.Neuron import IndividualNeuron
from ANNarchy.core.Synapse import Synapse
from ANNarchy.core.Dendrite import Dendrite
from ANNarchy.core.PopulationView import PopulationView
from ANNarchy.core import ConnectorMethods

from ANNarchy.intern.NetworkManager import NetworkManager
from ANNarchy.intern import Messages
from ANNarchy.intern.Profiler import Profiler
from ANNarchy.intern.ConfigManagement import ConfigManager, _check_paradigm
from ANNarchy.core.Constant import Constant

class Projection :
    """
    Projection between two populations.

    The object is returned by `Network.connect()` and should not be created directly:

    ```python
    proj = net.connect(pre=pop1, post=pop2, target="exc", synapse=STDP)
    ```

    The projection still has to be instantiated, by calling a connector method such as `all_to_all()` or `fixed_probability()`.

    If not specified, the default synapse only ensures linear synaptic transmission:

    * For rate-coded populations: ``psp = w * pre.r``
    * For spiking populations: ``g_target += w``

    :param pre: Pre-synaptic population (either its name or a ``Population`` object).
    :param post: Post-synaptic population (either its name or a ``Population`` object).
    :param target: Type of the connection.
    :param synapse: A `Synapse` instance.
    :param name: Unique name of the projection (optional, it defaults to ``proj0``, ``proj1``, etc).
    :param disable_omp: Especially for small- and mid-scale sparse spiking networks, the parallelization of spike propagation is not scalable and disabled by default. It can be re-enabled by setting this parameter to `False`.
    """

    def __init__(self, 
                 pre: str | Population, 
                 post: str | Population, 
                 target: str, 
                 synapse: Synapse = None, 
                 name:str = None, 
                 # Internal
                 disable_omp:bool = True, 
                 copied:bool = False,
                 net_id:int = 0):

        # Check if the network has already been compiled
        if NetworkManager().get_network(net_id).compiled and not copied:
            Messages._error('You cannot add a projection after the network has been compiled.')

        # Store net_id
        self.net_id = net_id

        # Store the pre and post synaptic populations
        # the user provide either a string or a population object
        # in case of string, we need to search for the corresponding object
        if isinstance(pre, str):
            for pop in NetworkManager().get_network(net_id).get_populations():
                if pop.name == pre:
                    self.pre = pop
        else:
            if isinstance(pre, IndividualNeuron):
                self.pre = pre.population
            else:
                self.pre = pre
                "Pre-synaptic population."

        if isinstance(post, str):
            for pop in NetworkManager().get_network(net_id).get_populations():
                if pop.name == post:
                    self.post = pop
        else:
            if isinstance(post, IndividualNeuron):
                self.post = post.population
            else:
                self.post = post
                "Post-synaptic population."

        # Store the arguments
        if isinstance(target, list) and len(target) == 1:
            self.target = target[0]
        else:
            self.target = target
            "Target."

        # Add the target(s) to the postsynaptic population
        if isinstance(self.target, list):
            for _target in self.target:
                self.post.targets.append(_target)
        else:
            self.post.targets.append(self.target)

        # check if a synapse description is attached
        if not synapse:
            # No synapse attached assume default synapse based on
            # presynaptic population.
            if self.pre.neuron_type.type == 'rate':
                from ANNarchy.models.Synapses import DefaultRateCodedSynapse
                self.synapse_type = DefaultRateCodedSynapse()
                self.synapse_type.type = 'rate'
            else:
                from ANNarchy.models.Synapses import DefaultSpikingSynapse
                self.synapse_type = DefaultSpikingSynapse()
                self.synapse_type.type = 'spike'

        elif inspect.isclass(synapse):
            self.synapse_type = synapse()
            self.synapse_type.type = self.pre.neuron_type.type
        else:
            self.synapse_type = copy.deepcopy(synapse)
            self.synapse_type.type = self.pre.neuron_type.type

        # Disable omp for spiking networks
        self.disable_omp = disable_omp

        # Analyse the parameters and variables
        self.synapse_type._analyse(self.net_id)

        # Create a default name
        self.id = NetworkManager().get_network(net_id)._add_projection(self)
        if name:
            self.name = name
        else:
            self.name = 'proj'+str(self.id)

        # Container for control/attribute states
        self.init = {}

        # Control-flow variables
        self.init["transmission"] = True
        self.init["axon_transmission"] = True
        self.init["update"] = True
        self.init["plasticity"] = True

        # Get a list of parameters and variables
        self.parameters = []
        "List of parameter names."
        for param in self.synapse_type.description['parameters']:
            self.parameters.append(param['name'])
            self.init[param['name']] = param['init']

        self.variables = []
        "List of variable names."
        for var in self.synapse_type.description['variables']:
            self.variables.append(var['name'])
            self.init[var['name']] = var['init']

        self.attributes = self.parameters + self.variables
        "List of attribute names (parameters + variables)."

        # Get a list of user-defined functions
        self.functions = [func['name'] for func in self.synapse_type.description['functions']]
        "List of functions defined by the synapse model."

        # Finalize initialization
        self.initialized = False

        # Cython instance
        self.cyInstance = None

        # Connectivity
        self._synapses = None
        self._connection_method = None
        self._connection_args = None
        self._connection_delay = None
        self._connector = None
        self._lil_connectivity = None

        # Default configuration for connectivity
        self._storage_format = "lil"
        self._storage_order = "post_to_pre"

        # If a single weight value is used
        self._single_constant_weight = False

        # Are random distribution used for weights/delays
        self.connector_weight_dist = None
        self.connector_delay_dist = None

        # Reporting
        self.connector_name = "Specific"
        self.connector_description = "Specific"

        # Overwritten by derived classes, to add
        # additional code
        self._specific_template = {}

        # Set to False by derived classes to prevent saving of
        # data, e. g. in case of weight-sharing projections
        self._saveable = True

        # To allow case-specific adjustment of parallelization
        # parameters, e. g. openMP schedule, we introduce a
        # dictionary read by the ProjectionGenerator.
        #
        # Will be overwritten either by inherited classes or
        # by an omp_config provided to the compile() method.
        self._omp_config = {
            #'psp_schedule': 'schedule(dynamic)'
        }

        # If set to true, the code generator is not allowed to
        # split the matrix. This will be the case for many
        # SpecificProjections defined by the user or is disabled
        # globally.
        if self.synapse_type.type == "rate":
            # Normally, the split should not be used for rate-coded models
            # but maybe there are cases where we want to enable it ...
            self._no_split_matrix = ConfigManager().get('disable_split_matrix', self.net_id)

            # If the number of elements is too small, the split
            # might not be efficient.
            if self.post.size < Global.OMP_MIN_NB_NEURONS:
                self._no_split_matrix = True

        else:
            # If the number of elements is too small, the split
            # might not be efficient.
            if self.post.size < Global.OMP_MIN_NB_NEURONS:
                self._no_split_matrix = True
            else:
                self._no_split_matrix = ConfigManager().get('disable_split_matrix', self.net_id)

        # In particular for spiking models, the parallelization on the
        # inner or outer loop can make a performance difference
        if self._no_split_matrix:
            # LIL and CSR are parallelized on inner loop
            # to prevent cost of atomic operations
            self._parallel_pattern = 'inner_loop'
        else:
            # splitted matrices are always parallelized on outer loop!
            self._parallel_pattern = 'outer_loop'

    ################################
    ## Connectivity methods
    ## 
    ## connect_xxx is defined for 4.x legacy
    ################################

    # All-to-all
    all_to_all = ConnectorMethods.connect_all_to_all
    connect_all_to_all = all_to_all

    # Fixed probability
    fixed_probability = ConnectorMethods.connect_fixed_probability
    connect_fixed_probability = fixed_probability

    # One-to-one
    one_to_one = ConnectorMethods.connect_one_to_one
    connect_one_to_one = one_to_one

    # Fixed number of pre neurons
    fixed_number_pre = ConnectorMethods.connect_fixed_number_pre
    connect_fixed_number_pre = fixed_number_pre

    # Fixed number of post neurons
    fixed_number_post = ConnectorMethods.connect_fixed_number_post
    connect_fixed_number_post = fixed_number_post

    # Gaussian and Difference-of-Gaussian (DoG)
    gaussian = ConnectorMethods.connect_gaussian
    connect_gaussian = gaussian
    dog = ConnectorMethods.connect_dog
    connect_dog = dog

    # From functions
    from_function = ConnectorMethods.connect_with_func
    connect_with_func = from_function

    # From matrices
    from_matrix = ConnectorMethods.connect_from_matrix
    connect_from_matrix = from_matrix
    from_matrix_market = ConnectorMethods.connect_from_matrix_market
    connect_from_matrix_market = from_matrix_market
    from_sparse = ConnectorMethods.connect_from_sparse
    connect_from_sparse = from_sparse

    # From file
    from_file = ConnectorMethods.connect_from_file
    connect_from_file = from_file

    # Loaders
    _load_from_matrix = ConnectorMethods._load_from_matrix
    _load_from_lil = ConnectorMethods._load_from_lil
    _load_from_sparse = ConnectorMethods._load_from_sparse

    def _copy(self, pre, post, net_id=None):
        "Returns a copy of the projection when creating networks.  Internal use only."
        
        copied_proj = Projection(
            pre=pre, 
            post=post, 
            target=self.target, 
            synapse=self.synapse_type, 
            name=self.name, 
            disable_omp=self.disable_omp, 
            copied=True, 
            net_id = self.net_id if net_id is None else net_id)

        # these flags are modified during connect_XXX called before Network()
        copied_proj._single_constant_weight = self._single_constant_weight
        copied_proj.connector_weight_dist = self.connector_weight_dist
        copied_proj.connector_delay_dist = self.connector_delay_dist
        copied_proj.connector_name = self.connector_name

        # Control flags for code generation (maybe modified by connect_XXX())
        copied_proj._storage_format = self._storage_format
        copied_proj._storage_order = self._storage_order
        copied_proj._no_split_matrix = self._no_split_matrix

        # Specific for OpenMP
        copied_proj._parallel_pattern = self._parallel_pattern
        copied_proj._no_split_matrix = self._no_split_matrix

        # for some projection types saving is not allowed (e. g. Convolution, Pooling)
        copied_proj._saveable = self._saveable

        # Additional configuration flags for sparse matrix formats
        if hasattr(self, "_bsr_tile_size"):
            copied_proj._bsr_tile_size = self._bsr_tile_size
        if hasattr(self, "_sell_block_size"):
            copied_proj._sell_block_size = self._sell_block_size

        return copied_proj

    def _generate(self):
        "Overriden by specific projections to generate the code"
        pass

    def _instantiate(self, module):
        """
        Instantiates the projection after compilation. The function should be
        called by Compiler._instantiate().

        :param module:  cython module (ANNarchyCore instance)
        """
        if Profiler().enabled:
            import time
            t1 = time.time()

        self.initialized = self._connect(module)

        if Profiler().enabled:
            t2 = time.time()
            Profiler().add_entry(t1, t2, "proj"+str(self.id), "instantiate")

    def _init_attributes(self):
        """
        Method used after compilation to initialize the attributes. The function
        should be called by Compiler._instantiate
        """
        for name, value in self.init.items():
            # The weights ('w') are already initialized by the _connect() method.
            if name in ['w']:
                continue

            if isinstance(value, Constant):
                self.__setattr__(name, value.value)
            elif isinstance(value, RandomDistribution): # The initial value of a variable is a random variable
                self.__setattr__(name, value)
            elif isinstance(value, str): # The initial value of a variable is a parameter
                self.__setattr__(name, self.__getattr__(value))
            else:
                self.__setattr__(name, value)

    def _connect(self, module):
        """
        Builds up dendrites either from list or dictionary. Called by instantiate().

        Returns True, if the connector was successfully instantiated. Potential errors are kept by  Python exceptions. If the Cython connector call fails (return False) the most likely reason is that there was not enough memory available.

        :param module:  cython module (ANNarchyCore instance)

        """
        # Local import to prevent circular import (HD: 28th June 2021)
        from ANNarchy.generator.Utils import cpp_connector_available

        # Sanity check
        if not self._connection_method:
            Messages._error('The projection between ' + self.pre.name + ' and ' + self.post.name + ' is declared but not connected.')

        # Debug printout
        if ConfigManager().get('verbose', self.net_id):
            print("  Initialize connectivity for Projection '"+self.name+"':")
            print("    pattern  :", self.connector_name)
            print("    arguments:", self._connection_args )

        # Instantiate the Cython wrapper
        if not self.cyInstance:
            cy_wrapper = getattr(module, 'proj'+str(self.id)+'_wrapper')
            self.cyInstance = cy_wrapper()

        # Check if there is a specialized CPP connector. No default connector -> initialize from LIL
        if not cpp_connector_available(self.connector_name, self._storage_format, self._storage_order, self.net_id):
            if not self._lil_connectivity:
                # Call the connector method (either cythonized or user-defined python method)
                synapses = self._connection_method(*((self.pre, self.post,) + self._connection_args))
                success = self.cyInstance.init_from_lil(synapses.post_rank, synapses.pre_rank, synapses.w, synapses.delay, synapses.requires_sorting)
            else:
                # LIL connectivity was built already by auto-tuning.
                success = self.cyInstance.init_from_lil(self._lil_connectivity.post_rank, self._lil_connectivity.pre_rank, self._lil_connectivity.w, self._lil_connectivity.delay, self._lil_connectivity.requires_sorting)
                del self._lil_connectivity      # Trigger destruction of the cython instance.
                self._lil_connectivity = None   # Otherwise this would retain until end of the simulations

            return success

        else:
            if ConfigManager().get('verbose', self.net_id):
                print("Use CPP-side implementation of", self.connector_name,"pattern for ProjStruct"+str(self.id))

            # all-to-all pattern
            if self.connector_name == "All-to-All":
                if isinstance(self._connection_args[0], RandomDistribution):
                    #some kind of distribution
                    w_dist_arg1, w_dist_arg2 = self._connection_args[0].get_cpp_args()
                else:
                    # constant
                    w_dist_arg1 = self._connection_args[0]
                    w_dist_arg2 = self._connection_args[0]

                if isinstance(self._connection_args[1], RandomDistribution):
                    #some kind of distribution
                    d_dist_arg1, d_dist_arg2 = self._connection_args[1].get_cpp_args()
                else:
                    # constant
                    d_dist_arg1 = self._connection_args[1]
                    d_dist_arg2 = self._connection_args[1]
                allow_self_connections = self._connection_args[2]

                return self.cyInstance.all_to_all(self.post.ranks, self.pre.ranks, w_dist_arg1, w_dist_arg2, d_dist_arg1, d_dist_arg2, allow_self_connections)

            # fixed probability pattern
            elif self.connector_name == "Random":
                p = self._connection_args[0]
                allow_self_connections = self._connection_args[3]
                if isinstance(self._connection_args[1], RandomDistribution):
                    #some kind of distribution
                    w_dist_arg1, w_dist_arg2 = self._connection_args[1].get_cpp_args()
                else:
                    # constant
                    w_dist_arg1 = self._connection_args[1]
                    w_dist_arg2 = self._connection_args[1]

                if isinstance(self._connection_args[2], RandomDistribution):
                    #some kind of distribution
                    d_dist_arg1, d_dist_arg2 = self._connection_args[2].get_cpp_args()
                else:
                    # constant
                    d_dist_arg1 = self._connection_args[2]
                    d_dist_arg2 = self._connection_args[2]

                return self.cyInstance.fixed_probability(self.post.ranks, self.pre.ranks, p, w_dist_arg1, w_dist_arg2, d_dist_arg1, d_dist_arg2, allow_self_connections)

            # fixed number pre prattern
            elif self.connector_name== "Random Convergent":
                number_nonzero = self._connection_args[0]
                if isinstance(self._connection_args[1], RandomDistribution):
                    #some kind of distribution
                    w_dist_arg1, w_dist_arg2 = self._connection_args[1].get_cpp_args()
                else:
                    # constant
                    w_dist_arg1 = self._connection_args[1]
                    w_dist_arg2 = self._connection_args[1]

                if isinstance(self._connection_args[2], RandomDistribution):
                    #some kind of distribution
                    d_dist_arg1, d_dist_arg2 = self._connection_args[2].get_cpp_args()
                else:
                    # constant
                    d_dist_arg1 = self._connection_args[2]
                    d_dist_arg2 = self._connection_args[2]

                return self.cyInstance.fixed_number_pre(self.post.ranks, self.pre.ranks, number_nonzero, w_dist_arg1, w_dist_arg2, d_dist_arg1, d_dist_arg2)

            else:
                # This should never happen ...
                Messages._error("No initialization for CPP-connector defined ...")

        # should be never reached ...
        return False

    def _store_connectivity(self, method, args, delay, storage_format=None, storage_order=None):
        """
        Store connectivity data. This function is called from cython_ext.Connectors module.
        """
        # No format specified for this projection by the user, so fall-back to Global setting
        if storage_format is None:
            if ConfigManager().get('sparse_matrix_format', self.net_id) == "default":
                if _check_paradigm("openmp", self.net_id):
                    storage_format = "lil"
                elif _check_paradigm("cuda", self.net_id):
                    storage_format = "csr"
                else:
                    raise NotImplementedError

            else:
                storage_format = ConfigManager().get('sparse_matrix_format', self.net_id)

        # No storage order specified for this projection by the user, so fall-back to Global setting
        if storage_order is None:
            storage_order = ConfigManager().get('sparse_matrix_storage_order', self.net_id)

        # Sanity checks
        if self._connection_method != None:
            Messages._warning("Projection ", self.name, " was already connected ... data will be overwritten.")

        # Store connectivity pattern parameters
        self._connection_method = method
        self._connection_args = args
        self._connection_delay = delay
        self._storage_format = storage_format
        self._storage_order = storage_order

        # The user selected nothing therefore we use the standard since ANNarchy 4.4.0
        if storage_format is None:
            self._storage_format = "lil"
        if storage_order is None:
            if storage_format == "auto":
                storage_order = "auto"
            else:
                self._storage_order = "post_to_pre"

        # The user selected automatic format selection using heuristics
        if storage_format == "auto":
            self._storage_format = self._automatic_format_selection()
        if storage_order == "auto":
            self._storage_order = self._automatic_order_selection()

        # Analyse the delay
        if isinstance(delay, (int, float)): # Uniform delay
            self.max_delay = round(delay/ConfigManager().get('dt', self.net_id))
            self.uniform_delay = round(delay/ConfigManager().get('dt', self.net_id))

        elif isinstance(delay, RandomDistribution): # Non-uniform delay
            self.uniform_delay = -1
            # Ensure no negative delays are generated
            if delay.min is None or delay.min < ConfigManager().get('dt', self.net_id):
                delay.min = ConfigManager().get('dt', self.net_id)
            # The user needs to provide a max in order to compute max_delay
            if delay.max is None:
                Messages._error('Projection connector: if you use a non-bounded random distribution for the delays (e.g. Normal), you need to set the max argument to limit the maximal delay.')

            self.max_delay = round(delay.max/ConfigManager().get('dt', self.net_id))

        elif isinstance(delay, (list, np.ndarray)): # connect_from_matrix/sparse
            if len(delay) > 0:
                self.uniform_delay = -1
                self.max_delay = round(max([max(l) for l in delay])/ConfigManager().get('dt', self.net_id))
            else: # list is empty, no delay
                self.max_delay = -1
                self.uniform_delay = -1

        else:
            Messages._error('Projection connector: delays are not valid!')

        # Transmit the max delay to the pre pop
        if isinstance(self.pre, PopulationView):
            self.pre.population.max_delay = max(self.max_delay, self.pre.population.max_delay)
        else:
            self.pre.max_delay = max(self.max_delay, self.pre.max_delay)

    def _automatic_format_selection(self):
        """
        We check some heuristics to select a specific format implemented as decision tree:

            - If the filling degree is high enough a full matrix representation might be better
            - if the average row length is below a threshold the ELLPACK-R might be better
            - if the average row length is higher than a threshold the CSR might be better

        HD (17th Jan. 2022): Currently structural plasticity is only usable with LIL. But one could also
                             apply it for dense matrices in the future. For CSR and in particular the ELL-
                             like formats the potential memory-reallocations make the structural plasticity
                             a costly operation.
        """
        # Local import to prevent circular dependency error
        from ANNarchy.intern.SpecificProjection import SpecificProjection

        # Connection pattern / Feature specific selection
        if ConfigManager().get('structural_plasticity', self.net_id):
            storage_format = "lil"

        elif isinstance(self, SpecificProjection):
            storage_format = "lil"

        elif self.connector_name == "All-to-All":
            storage_format = "dense"

        elif self.connector_name == "One-to-One":
            if _check_paradigm("cuda", self.net_id):
                storage_format = "csr"
            else:
                storage_format = "lil"

        else:
            if self.synapse_type.type == "spike":
                # we need to build up the matrix to analyze
                self._lil_connectivity = self._connection_method(*((self.pre, self.post,) + self._connection_args))

                # get the decision parameter
                density = float(self._lil_connectivity.nb_synapses) / float(self.pre.size * self.post.size)
                if density >= 0.6:
                    storage_format = "dense"
                else:
                    storage_format = "csr"

            else:
                # we need to build up the matrix to analyze
                self._lil_connectivity = self._connection_method(*((self.pre, self.post,) + self._connection_args))

                # get the decision parameter
                density = float(self._lil_connectivity.nb_synapses) / float(self.pre.size * self.post.size)
                avg_nnz_per_row, _, _, _ = self._lil_connectivity.compute_average_row_length()

                # heuristic decision tree
                if density >= 0.6:
                    storage_format = "dense"
                else:
                    if _check_paradigm("cuda", self.net_id):
                        if avg_nnz_per_row <= 128:
                            if self.synapse_type.description['plasticity']:
                                storage_format = "ellr"
                            else:
                                storage_format = "sell"
                        else:
                            storage_format = "csr"
                    else:
                        storage_format = "csr"

        Messages._info("Automatic format selection for", self.name, ":", storage_format)
        return storage_format

    def _automatic_order_selection(self):
        """
        Contrary to the matrix format, the decision for the matrix order is majorly dependent on
        the synapse type.
        """
        if self.synapse_type.type == "rate":
            storage_order = "post_to_pre"
        else:
            from ANNarchy.generator.Utils import sort_odes

            # Check if synapse-related ODEs are continous
            odes = sort_odes(self.synapse_type.description, 'local')
            continuous_odes = False if odes == [] else True

            if 'psp' in  self.synapse_type.description.keys():
                # continuous signal transmission should always be post-to-pre
                storage_order = "post_to_pre"
            elif continuous_odes:
                # continuous evaluated ODEs will dominate the computation therefore should be post-to-pre
                storage_order = "post_to_pre"
            else:
                # pre-to-post is not implemented for all formats
                if self._storage_format in ["dense", "csr"]:
                    storage_order = "pre_to_post"
                else:
                    storage_order = "post_to_pre"

        Messages._info("Automatic matrix order selection for", self.name, ":", storage_order)
        return storage_order

    def _has_single_weight(self):
        "If a single weight should be generated instead of a LIL"
        is_cpu = ConfigManager().get('paradigm', self.net_id)=="openmp"
        has_constant_weight = self._single_constant_weight
        not_dense = not (self._storage_format == "dense")
        no_structural_plasticity = not ConfigManager().get('structural_plasticity', self.net_id)
        no_synaptic_plasticity = not self.synapse_type.description['plasticity']

        return has_constant_weight and no_structural_plasticity and no_synaptic_plasticity and is_cpu and not_dense

    def reset(self, attributes=-1, synapses=False):
        """
        Resets all parameters and variables of the projection to their initial value (before the call to compile()).

        :param attributes: list of attributes (parameter or variable) which should be reinitialized. Default: all attributes (-1).
        :param synapses: defines if the weights and delays should also be recreated. Default: False
        """
        if attributes == -1:
            attributes = self.attributes

        if synapses:
            # destroy the previous C++ content
            self._clear()
            # call the init connectivity again
            self.initialized = self._connect(None)

        for var in attributes:
            # Skip w
            if var=='w':
                continue
            # check it exists
            if not var in self.attributes:
                Messages._warning("Projection.reset():", var, "is not an attribute of the population, won't reset.")
                continue
            # Set the value
            try:
                self.__setattr__(var, self.init[var])
            except Exception as e:
                Messages._print(e)
                Messages._warning("Projection.reset(): something went wrong while resetting", var)

    ################################
    ## Dendrite access
    ################################
    @property
    def size(self):
        "Number of post-synaptic neurons receiving synapses."
        if self.cyInstance is None:
            Messages._warning("Access 'size or len()' attribute of a Projection is only valid after compile()")
            return 0

        return len(self.cyInstance.post_rank())

    def __len__(self):
        # Number of postsynaptic neurons receiving synapses in this projection.
        return self.size

    @property
    def nb_synapses(self):
        "Total number of synapses in the projection."
        if self.cyInstance is None:
            Messages._warning("Access 'nb_synapses' attribute of a Projection is only valid after compile()")
            return 0
        return self.cyInstance.nb_synapses()

    @property
    def nb_synapses_per_dendrite(self):
        "Total number of synapses for each dendrite as a list."
        if self.cyInstance is None:
            Messages._warning("Access 'nb_synapses_per_dendrite' attribute of a Projection is only valid after compile()")
            return []
        return [self.cyInstance.dendrite_size(n) for n in range(self.size)]

    def _nb_efferent_synapses(self):
        """
        Returns the number of efferent connections. Intended only for spiking models.
        """
        if self.cyInstance is None:
             Messages._warning("Access 'nb_efferent_synapses()' of a Projection is only valid after compile()")
             return None
        if self.synapse_type.type == "rate":
            Messages._error("Projection.nb_efferent_synapses() is not available for rate-coded projections.")

        return self.cyInstance.nb_efferent_synapses()

    @property
    def post_ranks(self):
        "List of ranks of post-synaptic neurons that receive connections. Read-only."
        if self.cyInstance is None:
             Messages._warning("Access 'post_ranks' attribute of a Projection is only valid after compile()")
             return None
        
        return self.cyInstance.post_rank()
    
    @property
    def pre_ranks(self):
        "List of lists of pre-synaptic ranks, for each post-synaptic neuron. Read-only."
        if self.cyInstance is None:
             Messages._warning("Access 'pre_ranks' attribute of a Projection is only valid after compile()")
             return None
        
        return self.cyInstance.pre_ranks()

    @property
    def dendrites(self) -> Iterator[Dendrite]:
        """
        Iteratively returns the dendrites corresponding to this projection.
        """
        for idx, n in enumerate(self.post_ranks):
            yield Dendrite(self, n, idx)

    def dendrite(self, post:int) -> Dendrite:
        """
        Returns the dendrite of a postsynaptic neuron according to its rank.

        :param post: can be either the rank or the coordinates of the post-synaptic neuron.
        """
        if not self.initialized:
            Messages._error('dendrites can only be accessed after compilation.')

        if isinstance(post, (int, np.int64, np.int32)):
            rank = post
        else:
            rank = self.post.rank_from_coordinates(post)

        if rank in self.post_ranks:
            return Dendrite(self, rank, self.post_ranks.index(rank))
        else:
            Messages._error(" The neuron of rank "+ str(rank) + " has no dendrite in this projection.", exit=True)


    def synapse(self, pre:int, post:int) -> "IndividualSynapse":
        """
        Returns the synapse between a pre- and a post-synaptic neuron if it exists, `None` otherwise.

        :param pre: rank of the pre-synaptic neuron.
        :param post: rank of the post-synaptic neuron.
        """
        if not isinstance(pre, (int, np.int64, np.int32)) or not isinstance(post, (int, np.int64, np.int32)):
            Messages._error('Projection.synapse() only accepts ranks for the pre and post neurons.')

        return self.dendrite(post).synapse(pre)


    # Iterators
    def __getitem__(self, *args, **kwds):
        # Returns dendrite of the given position in the postsynaptic population.
        # If only one argument is given, it is a rank. If it is a tuple, it is coordinates.

        if len(args) == 1:
            return self.dendrite(args[0])
        return self.dendrite(args)

    def __iter__(self) -> Iterator[Dendrite]:
        # Returns iteratively each dendrite in the population in ascending postsynaptic rank order.
        for idx, n in enumerate(self.post_ranks):
            yield Dendrite(self, n, idx)

    ################################
    ## Access to attributes
    ################################
    def get(self, name):
        """
        Returns a list of parameters/variables values for each dendrite in the projection.

        The list will have the same length as the number of actual dendrites (self.size), so it can be smaller than the size of the postsynaptic population. Use self.post_ranks to indice it.

        :param name: the name of the parameter or variable
        """
        return self.__getattr__(name)

    def set(self, value):
        """
        Sets the parameters/variables values for each dendrite in the projection.

        For parameters, you can provide:

        * a single value, which will be the same for all dendrites.

        * a list or 1D numpy array of the same length as the number of actual dendrites (self.size).

        For variables, you can provide:

        * a single value, which will be the same for all synapses of all dendrites.

        * a list or 1D numpy array of the same length as the number of actual dendrites (self.size). The synapses of each postsynaptic neuron will take the same value.

        **Warning:** it is not possible to set different values to each synapse using this method. One should iterate over the dendrites:

        ```python
        for dendrite in proj.dendrites:
            dendrite.w = np.ones(dendrite.size)
        ```

        :param value: a dictionary with the name of the parameter/variable as key.

        """

        for name, val in value.items():
            self.__setattr__(name, val)

    def __getattr__(self, name):
        # Method called when accessing an attribute.
        if name == 'initialized' or not hasattr(self, 'initialized'): # Before the end of the constructor
            return object.__getattribute__(self, name)
        elif hasattr(self, 'attributes'):
            if name in ['plasticity', 'axon_transmission', 'transmission', 'update']:
                return self._get_flag(name)
            if name in ['delay']:
                return self._get_delay()
            if name in self.attributes:
                if not self.initialized:
                    return self.init[name]
                else:
                    return self._get_cython_attribute( name )
            elif name in self.functions:
                return self._function(name)
            else:
                return object.__getattribute__(self, name)
        return object.__getattribute__(self, name)

    def __setattr__(self, name, value):
        # Method called when setting an attribute.
        if name == 'initialized' or not hasattr(self, 'initialized'): # Before the end of the constructor
            object.__setattr__(self, name, value)
        elif hasattr(self, 'attributes'):
            if name in ['plasticity', 'axon_transmission', 'transmission', 'update']:
                self._set_flag(name, bool(value))
                return
            if name in ['delay']:
                self._set_delay(value)
                return
            if name in self.attributes:
                if not self.initialized:
                    self.init[name] = value
                else:
                    self._set_cython_attribute(name, value)
            else:
                object.__setattr__(self, name, value)
        else:
            object.__setattr__(self, name, value)

    def _get_cython_attribute(self, attribute):
        """
        Returns the value of the given attribute for all neurons in the population,
        as a list of lists having the same geometry as the population if it is local.

        :param attribute: a string representing the variables's name.

        """
        # Determine C++ data type
        ctype = self._get_attribute_cpp_type(attribute=attribute)

        if attribute == "w" and self._has_single_weight():
            return getattr(self.cyInstance, 'get_global_attribute_'+ctype)(attribute)
        elif attribute in self.synapse_type.description["local"]:
            return getattr(self.cyInstance, "get_local_attribute_all_"+ctype)(attribute)
        elif attribute in self.synapse_type.description["semiglobal"]:
            return getattr(self.cyInstance, "get_semiglobal_attribute_all_"+ctype)(attribute)
        elif attribute in self.synapse_type.description["global"]:
            return getattr(self.cyInstance, "get_global_attribute_"+ctype)(attribute)
        else:
            raise AttributeError("Attribute '"+attribute+"' does not seem to belong to Projection '"+self.name+"'.")

    def _set_cython_attribute(self, attribute, value):
        """
        Sets the value of the given attribute for all post-synaptic neurons in the projection,
        as a NumPy array having the same geometry as the population if it is local.

        :param attribute: a string representing the variables's name.
        :param value: the value it should take.

        """
        # Determine C++ data type
        ctype = self._get_attribute_cpp_type(attribute=attribute)

        # Convert np.arrays into lists/constants for better iteration
        if isinstance(value, np.ndarray):
            if np.ndim(value) == 0:
                value = float(value)
            else:
                value = value.tolist()

        # A list is given
        if isinstance(value, list):
            if len(value) == len(self.post_ranks):
                if attribute in self.synapse_type.description['local']:
                    # get old value
                    tmp = getattr(self.cyInstance, "get_local_attribute_all_"+ctype)(attribute)
                    for idx, n in enumerate(self.post_ranks):
                        if not len(value[idx]) == self.cyInstance.dendrite_size(idx):
                            Messages._error('The post-synaptic neuron ' + str(n) + ' of population ' + str(self.post.id) + ' receives '+ str(self.cyInstance.dendrite_size(idx))+ ' synapses and not ' + str(len(value[idx])) + '.')
                        # update single row
                        tmp[idx] = value[idx]
                    # write to C++ core
                    getattr(self.cyInstance, "set_local_attribute_all_"+ctype)(attribute, tmp)
                elif attribute in self.synapse_type.description['semiglobal']:
                    getattr(self.cyInstance, "set_semiglobal_attribute_all_"+ctype)(attribute, value)
                else:
                    Messages._error('The parameter', attribute, 'is global to the population, cannot assign a list.')
            else:
                Messages._error('The projection has', self.size, 'post-synaptic neurons, the list must have the same size.')

        # A Random Distribution is given
        elif isinstance(value, RandomDistribution):
            if attribute == "w" and self._has_single_weight():
                setattr(self.cyInstance, attribute, value.get_values(1))
            elif attribute in self.synapse_type.description['local']:
                # get old value
                tmp = getattr(self.cyInstance, "get_local_attribute_all_"+ctype)(attribute)
                # update
                for idx, n in enumerate(self.post_ranks):
                    tmp[idx] = value.get_values(self.cyInstance.dendrite_size(idx))
                # write to C++ core
                getattr(self.cyInstance, "set_local_attribute_all_"+ctype)(attribute, tmp)
            elif attribute in self.synapse_type.description['semiglobal']:
                getattr(self.cyInstance, "set_semiglobal_attribute_all_"+ctype)(attribute, value.get_values(len(self.post_ranks)))
            elif attribute in self.synapse_type.description['global']:
                getattr(self.cyInstance, "set_global_attribute_"+ctype)(attribute, value.get_values(1))

        # A single value is given
        else:
            if attribute == "w" and self._has_single_weight():
                getattr(self.cyInstance, 'set_global_attribute_'+ctype)(attribute, value)
            
            elif attribute in self.synapse_type.description['local']:
                # get old value
                tmp = getattr(self.cyInstance, "get_local_attribute_all_"+ctype)(attribute)
                # update
                for idx, n in enumerate(self.post_ranks):
                    tmp[idx] = [value for _ in range(self.cyInstance.dendrite_size(idx))]
                # write to C++ core
                getattr(self.cyInstance, "set_local_attribute_all_"+ctype)(attribute, tmp)
            
            elif attribute in self.synapse_type.description['semiglobal']:
                getattr(self.cyInstance, 'set_semiglobal_attribute_all_'+ctype)(attribute, 
                        [value for _ in range(len(self.post_ranks))])
            
            else:
                getattr(self.cyInstance, 'set_global_attribute_'+ctype)(attribute, value)

    def _get_attribute_cpp_type(self, attribute):
        """
        Determine C++ data type for a given attribute
        """
        ctype = None
        for var in self.synapse_type.description['variables']+self.synapse_type.description['parameters']:
            if var['name'] == attribute:
                ctype = var['ctype']

        return ctype

    def _get_flag(self, attribute):
        "control flow flags such as learning, transmission"
        if self.cyInstance is not None:
            return getattr(self.cyInstance, '_'+attribute)
        else:
            return self.init[attribute]

    def _set_flag(self, attribute, value):
        "control flow flags such as learning, transmission"
        if self.cyInstance is not None:
            setattr(self.cyInstance, '_'+attribute, value)
        else:
            self.init[attribute] = value


    ################################
    ## Access to delays
    ################################
    def _get_delay(self):
        if not hasattr(self.cyInstance, 'get_delay'):
            if self.max_delay <= 1 :
                return ConfigManager().get('dt', self.net_id)
        elif self.uniform_delay != -1:
            return self.uniform_delay * ConfigManager().get('dt', self.net_id)
        else:
            return [[pre * ConfigManager().get('dt', self.net_id) for pre in post] for post in self.cyInstance.get_delay()]

    def _set_delay(self, value):

        if self.cyInstance: # After compile()
            if not hasattr(self.cyInstance, 'get_delay'):
                if self.max_delay <= 1 and value != ConfigManager().get('dt', self.net_id):
                    Messages._error("set_delay: the projection was instantiated without delays, it is too late to create them...")

            elif self.uniform_delay != -1:
                if isinstance(value, np.ndarray):
                    if value.ndim > 0:
                        Messages._error("set_delay: the projection was instantiated with uniform delays, it is too late to load non-uniform values...")
                    else:
                        value = max(1, round(float(value)/ConfigManager().get('dt', self.net_id)))
                elif isinstance(value, (float, int)):
                    value = max(1, round(float(value)/ConfigManager().get('dt', self.net_id)))
                else:
                    Messages._error("set_delay: only float, int or np.array values are possible.")

                # The new max_delay is higher than before
                if value > self.max_delay:
                    self.max_delay = value
                    self.uniform_delay = value
                    self.cyInstance.set_delay(value)
                    if isinstance(self.pre, PopulationView):
                        self.pre.population.max_delay = max(self.max_delay, self.pre.population.max_delay)
                        self.pre.population.cyInstance.update_max_delay(self.pre.population.max_delay)
                    else:
                        self.pre.max_delay = max(self.max_delay, self.pre.max_delay)
                        self.pre.cyInstance.update_max_delay(self.pre.max_delay)
                    return
                else:
                    self.uniform_delay = value
                    self.cyInstance.set_delay(value)

            else: # variable delays
                if not isinstance(value, (np.ndarray, list)):
                    Messages._error("set_delay with variable delays: you must provide a list of lists of exactly the same size as before.")

                # Check the number of delays
                nb_values = sum([len(s) for s in value])
                if nb_values != self.nb_synapses:
                    Messages._error("set_delay with variable delays: the sizes do not match. You have to provide one value for each existing synapse.")
                if len(value) != len(self.post_ranks):
                    Messages._error("set_delay with variable delays: the sizes do not match. You have to provide one value for each existing synapse.")

                # Convert to steps
                if isinstance(value, np.ndarray):
                    delays = [[max(1, round(value[i, j]/ConfigManager().get('dt', self.net_id))) for j in range(value.shape[1])] for i in range(value.shape[0])]
                else:
                    delays = [[max(1, round(v/ConfigManager().get('dt', self.net_id))) for v in c] for c in value]

                # Max delay
                max_delay = max([max(l) for l in delays])

                if max_delay > self.max_delay:
                    self.max_delay = max_delay

                    # Send the max delay to the pre population
                    if isinstance(self.pre, PopulationView):
                        self.pre.population.max_delay = max(self.max_delay, self.pre.population.max_delay)
                        self.pre.population.cyInstance.update_max_delay(self.pre.population.max_delay)
                    else:
                        self.pre.max_delay = max(self.max_delay, self.pre.max_delay)
                        self.pre.cyInstance.update_max_delay(self.pre.max_delay)

                # Send the new values to the projection
                self.cyInstance.set_delay(delays)

                # Update ring buffers (if there exist)
                if self.synapse_type == "spike":
                    self.cyInstance.update_max_delay(self.max_delay)

        else: # before compile()
            Messages._error("set_delay before compile(): not implemented yet.")


    ################################
    ## Access to functions
    ################################
    def _function(self, name):
        "Access a user defined function"
        if not self.initialized:
            Messages._warning('the network is not compiled yet, cannot access the function ' + name)
            return
        
        # Get the C++ function
        cpp_function = getattr(self.cyInstance, name)

        # One argument
        def apply(*args):
            return list(map(cpp_function, *args))
        return apply

    ################################
    ## Learning flags
    ################################
    def enable_learning(self, period:float=None, offset:float=None) -> None:
        """
        Enables learning for all the synapses of this projection.

        For example, providing the following parameters at time 10 ms:

        ```python
        proj.enable_learning(period=10., offset=5.)
        ```

        would call the updating methods at times 15, 25, 35, etc...

        The default behaviour is that the synaptic variables are updated at each time step. The parameters must be multiple of ``dt``.

        :param period: determines how often the synaptic variables will be updated.
        :param offset: determines the offset at which the synaptic variables will be updated relative to the current time.

        """
        # Check arguments
        if not period is None and not offset is None:
            if offset >= period:
                Messages._error('enable_learning(): the offset must be smaller than the period.')

        if period is None and not offset is None:
            Messages._error('enable_learning(): if you define an offset, you have to define a period.')

        try:
            self.cyInstance._update = True
            self.cyInstance._plasticity = True
            if period != None:
                self.cyInstance._update_period = int(period/ConfigManager().get('dt', self.net_id))
            else:
                self.cyInstance._update_period = int(1)
                period = ConfigManager().get('dt', self.net_id)
            if offset != None:
                relative_offset = Global.get_time() % period + offset
                self.cyInstance._update_offset = int(int(relative_offset%period)/ConfigManager().get('dt', self.net_id))
            else:
                self.cyInstance._update_offset = int(0)
        except:
            Messages._warning('Enable_learning() is only possible after compile()')

    def disable_learning(self) -> None:
        """
        Disables learning for all synapses of this projection.

        The effect depends on the rate-coded or spiking nature of the projection:

        * **Rate-coded**: the updating of all synaptic variables is disabled (including the weights ``w``). This is equivalent to ``proj.update = False``.

        * **Spiking**: the updating of the weights ``w`` is disabled, but all other variables are updated. This is equivalent to ``proj.plasticity = False``.

        This method is useful when performing some tests on a trained network without messing with the learned weights.
        """
        try:
            if self.synapse_type.type == 'rate':
                self.cyInstance._update = False
            else:
                self.cyInstance._plasticity = False
        except:
            Messages._warning('disabling learning is only possible after compile().')


    ################################
    ## Methods on connectivity matrix
    ################################

    def save_connectivity(self, filename):
        """
        Saves the connectivity of the projection into a file.

        Only the connectivity matrix, the weights and delays are saved, not the other synaptic variables (use `save()` if you want these).

        ```python
        filename = 'data.npz'
        proj.save_connectivity(filename)
        ```

        The generated data can be used to create a projection in another network:

        ```python
        proj.from_file(filename)
        ```

        * If the file name is '.npz', the data will be saved and compressed using `np.savez_compressed` (recommended).
        * If the file name ends with '.gz', the data will be pickled into a binary file and compressed using gzip.
        * If the file name is '.mat', the data will be saved as a Matlab 7.2 file. Scipy must be installed.
        * Otherwise, the data will be pickled into a simple binary text file using pickle.

        :param filename: file name, may contain relative or absolute path.

        """
        # Check that the network is compiled
        if not self.initialized:
            Messages._error('save_connectivity(): the network has not been compiled yet.')
            return

        # Check if the repertory exist
        (path, fname) = os.path.split(filename)

        if not path == '':
            if not os.path.isdir(path):
                Messages._print('Creating folder', path)
                os.mkdir(path)

        extension = os.path.splitext(fname)[1]

        # Gathering the data
        data = {
            'name': self.name,
            'post_ranks': self.post_ranks,
            'pre_ranks': np.array(self.cyInstance.pre_ranks(), dtype=object),
            'w': np.array(self.w, dtype=object),
            'delay': np.array(self._get_delay(), dtype=object) if hasattr(self.cyInstance, 'get_delay') else None,
            'max_delay': self.max_delay,
            'uniform_delay': self.uniform_delay,
            'size': self.size,
            'nb_synapses': self.cyInstance.nb_synapses()
        }

        # Save the data
        if extension == '.gz':
            Messages._print("Saving connectivity in gunzipped binary format...")
            try:
                import gzip
            except ImportError:
                Messages._error('gzip is not installed.')
                return
            with gzip.open(filename, mode = 'wb') as w_file:
                try:
                    pickle.dump(data, w_file, protocol=pickle.HIGHEST_PROTOCOL)
                except Exception as e:
                    Messages._print('Error while saving in gzipped binary format.')
                    Messages._print(e)
                    return

        elif extension == '.npz':
            Messages._print("Saving connectivity in Numpy format...")
            np.savez_compressed(filename, **data )

        elif extension == '.mat':
            Messages._print("Saving connectivity in Matlab format...")
            if data['delay'] is None:
                data['delay'] = 0
            try:
                import scipy.io as sio
                sio.savemat(filename, data)
            except Exception as e:
                Messages._error('Error while saving in Matlab format.')
                Messages._print(e)
                return

        else:
            Messages._print("Saving connectivity in text format...")
            # save in Pythons pickle format
            with open(filename, mode = 'wb') as w_file:
                try:
                    pickle.dump(data, w_file, protocol=pickle.HIGHEST_PROTOCOL)
                except Exception as e:
                    Messages._print('Error while saving in text format.')
                    Messages._print(e)
                    return
            return

    def connectivity_matrix(self, fill:float=0.0) -> np.ndarray:
        """
        Returns a dense connectivity matrix (2D Numpy array) representing the connectivity matrix between the pre- and post-populations.

        The first index of the matrix represents post-synaptic neurons, the second the pre-synaptic ones.

        If PopulationViews were used for creating the projection, the matrix is expanded to the whole populations by default.

        :param fill: value to put in the matrix when there is no connection (default: 0.0).
        """
        if not self.initialized:
            Messages._error('The connectivity matrix can only be accessed after compilation')

        # get correct dimensions for dense matrix
        if isinstance(self.pre, PopulationView):
            size_pre = self.pre.population.size
        else:
            size_pre = self.pre.size
        if isinstance(self.post, PopulationView):
            size_post = self.post.population.size
        else:
            size_post = self.post.size

        # create empty dense matrix with default values
        res = np.ones((size_post, size_pre)) * fill

        # fill row-by-row with real values
        for rank in self.post_ranks:
            # row-rank
            idx =  self.post_ranks.index(rank)
            # pre-ranks
            preranks = self.cyInstance.pre_rank(idx)
            # get the values
            if "w" in self.synapse_type.description['local'] and (not self._has_single_weight()):
                w = getattr(self.cyInstance, "get_local_attribute_row_"+ConfigManager().get('precision', self.net_id))("w", idx)
            elif "w" in self.synapse_type.description['semiglobal']:
                w = getattr(self.cyInstance, "get_semiglobal_attribute_"+ConfigManager().get('precision', self.net_id))("w", idx)*np.ones(self.cyInstance.dendrite_size(idx))
            else:
                w = getattr(self.cyInstance, "get_global_attribute_"+ConfigManager().get('precision', self.net_id))("w")*np.ones(self.cyInstance.dendrite_size(idx))
            res[rank, preranks] = w
        return res

    def receptive_fields(self, variable:str='w', in_post_geometry:bool =True) -> np.ndarray:
        """
        Gathers all receptive fields within this projection.

        The method only works when the pre- and post-synaptic populations have a 2d geometry. The function
        concatenates all receptive fields of the `post` population into a 2d array.

        :param variable: Name of the variable.
        :param in_post_geometry: If False, the data will be plotted as square grid.
        """
        if in_post_geometry:
            x_size = self.post.geometry[1]
            y_size = self.post.geometry[0]
        else:
            x_size = int( math.floor(math.sqrt(self.post.size)) )
            y_size = int( math.ceil(math.sqrt(self.post.size)) )


        def get_rf(post_rank):
            try:
                lil_idx = self.post_ranks.index(post_rank)

                res = np.zeros( self.pre.size )
                pre_ranks = self.cyInstance.pre_rank(lil_idx)
                data = getattr(self.cyInstance, "get_local_attribute_row_"+ConfigManager().get('precision', self.net_id))(variable, lil_idx)

                if len(res) == len(pre_ranks):
                    res[pre_ranks] = data

                return res.reshape(self.pre.geometry)
            except ValueError:
                # post_rank is not listed
                return np.zeros(self.pre.geometry)

        res = np.zeros((1, x_size*self.pre.geometry[1]))
        for y in range ( y_size ):
            row = np.concatenate(  [ get_rf(self.post.rank_from_coordinates( (y, x) ) ) for x in range ( x_size ) ], axis = 1)
            res = np.concatenate((res, row))

        return res


    ################################
    ## Save/load methods
    ################################

    def _data(self):
        "Method gathering all info about the projection when calling save()"

        if not self.initialized:
            Messages._error('save(): the network has not been compiled yet.')

        desc = {}
        desc['name'] = self.name
        desc['pre'] = self.pre.name
        desc['post'] = self.post.name
        desc['target'] = self.target
        desc['post_ranks'] = self.post_ranks
        desc['attributes'] = self.attributes
        desc['parameters'] = self.parameters
        desc['variables'] = self.variables
        desc['delays'] = self._get_delay()

        # Determine if we have varying number of elements per row
        # based on the pre-synaptic ranks
        pre_ranks = self.cyInstance.pre_ranks()
        dend_size = len(pre_ranks[0])
        ragged_list = False
        for i in range(1, len(pre_ranks)):
            if len(pre_ranks[i]) != dend_size:
                ragged_list = True
                break

        # Save pre_ranks
        if ragged_list:
            desc['pre_ranks'] = np.array(self.cyInstance.pre_ranks(), dtype=object)
        else:
            desc['pre_ranks'] = np.array(self.cyInstance.pre_ranks())

        # Attributes to save
        attributes = self.attributes
        if not 'w' in self.attributes:
            attributes.append('w')

        # Save all attributes
        for var in attributes:
            try:
                ctype = self._get_attribute_cpp_type(var)
                
                if var == "w" and self._has_single_weight():
                    desc[var] = getattr(self.cyInstance, 'get_global_attribute_' + ctype)("w")

                elif var in self.synapse_type.description['local']:
                    if ragged_list:
                        desc[var] = np.array(getattr(self.cyInstance, 'get_local_attribute_all_' + ctype)(var), dtype=object)
                    else:
                        desc[var] = getattr(self.cyInstance, 'get_local_attribute_all_' + ctype)(var)
                
                elif var in self.synapse_type.description['semiglobal']:
                    desc[var] = getattr(self.cyInstance, 'get_semiglobal_attribute_all_' + ctype)(var)
                
                else:
                    desc[var] = getattr(self.cyInstance, 'get_global_attribute_' + ctype)(var) # linear array or single constant

            except Exception as e:
                Messages._print(e)
                Messages._warning('Can not save the attribute ' + var + ' in the projection (name='+str(self.name)+').')

        return desc

    def save(self, filename:str):
        """
        Saves all information about the projection (connectivity, current value of parameters and variables) into a file.

        * If the file name is '.npz', the data will be saved and compressed using `np.savez_compressed` (recommended).

        * If the file name ends with '.gz', the data will be pickled into a binary file and compressed using gzip.

        * If the file name is '.mat', the data will be saved as a Matlab 7.2 file. Scipy must be installed.

        * Otherwise, the data will be pickled into a simple binary text file using pickle.

        **Warning:** the '.mat' data will not be loadable by ANNarchy, it is only for external analysis purpose.

        Example:

        ```python
        proj.save('proj1.npz')
        proj.save('proj1.txt')
        proj.save('proj1.txt.gz')
        proj.save('proj1.mat')
        ```

        :param filename: file name, may contain relative or absolute path.
        """
        from ANNarchy.core.IO import _save_data
        _save_data(filename, self._data())


    def load(self, filename:str, pickle_encoding:str=None)  -> None:
        """
        Loads the saved state of the projection by `Projection.save()`.

        Warning: Matlab data can not be loaded.

        Example:

        ```python
        proj.load('proj1.npz')
        proj.load('proj1.txt')
        proj.load('proj1.txt.gz')
        ```

        :param filename: the file name with relative or absolute path.
        :param pickle_encoding: What encoding to use when reading Python 2 strings. Only useful when loading Python 2 generated pickled files in Python 3, which includes npy/npz files containing object arrays. Values other than `latin1`, `ASCII`, and `bytes` are not allowed, as they can corrupt numerical data. 
        """
        from ANNarchy.core.IO import _load_connectivity_data
        self._load_proj_data(_load_connectivity_data(filename, pickle_encoding))


    def _load_proj_data(self, desc):
        """
        Updates the projection with the stored data set.
        """
        # Sanity check
        if desc is None:
            # _load_proj should have printed an error message
            return

        # If it's not saveable there is nothing to load
        if not self._saveable:
            return

        # Check deprecation
        if not 'attributes' in desc.keys():
            Messages._error('The file was saved using a deprecated version of ANNarchy.')
            return
        if 'dendrites' in desc: # Saved before 4.5.3
            Messages._error("The file was saved using a deprecated version of ANNarchy.")
            return

        # Check row-sorting
        last_pr = -1
        requires_sorting = False
        for pr in desc['post_ranks']:
            if pr < last_pr:
                if requires_sorting == False:
                    requires_sorting = True
            last_pr = pr

        # If the post ranks and/or pre-ranks have changed, overwrite
        connectivity_changed=False
        # check the post-ranks
        if 'post_ranks' in desc and not np.all((desc['post_ranks']) == self.post_ranks):
            connectivity_changed=True
        # pre-ranks are stored as two-dimensional structure, however, dependent on the length
        # of inner vectors we have two cases: all equal-lengthed (two-dim matrix) or varying (ragged array)
        if 'pre_ranks' in desc:
            current_pre_ranks = np.array(self.cyInstance.pre_ranks(), dtype=object)

            # one array is ragged the other two-dimensional
            if desc['pre_ranks'].ndim != current_pre_ranks.ndim:
                connectivity_changed = True

            # both matrices are two-dimensional
            elif desc['pre_ranks'].shape != current_pre_ranks.shape:
                connectivity_changed = True

            # compare two ragged arrays
            elif not np.all((desc['pre_ranks']) == current_pre_ranks):
                connectivity_changed = True

        # Synaptic weights
        weights = desc["w"]

        # Delays can be either uniform (int, float) or non-uniform (np.ndarray).
        # HD (30th May 2022):
        #   Unfortunately, the storage of constants changed over the time. At the
        #   end of this code block, we should have either a single constant or a
        #   numpy nd-array
        delays = 0
        if 'delays' in desc:
            delays = desc['delays']

            if isinstance(delays, (float, int)):
                # will be handled below
                pass

            elif isinstance(delays, np.ndarray):
                # constants are stored as 0-darray
                if delays.ndim == 0:
                    # transform into single float
                    delays = float(delays)
                else:
                    # nothing to do as it is numpy nd-array
                    pass

            else:
                # ragged list to nd-array
                delays = np.array(delays, dtype=object)

        if not requires_sorting:
            # Some patterns like fixed_number_pre/post or fixed_probability change the
            # connectivity. If this is not the case, we can simply set the values.
            if connectivity_changed:
                # (re-)initialize connectivity
                if isinstance(delays, (float, int)):
                    delays = [[delays]] # wrapper expects list from list

                self.cyInstance.init_from_lil(desc['post_ranks'], desc['pre_ranks'], weights, delays, requires_sorting)
            else:
                # set weights
                self._set_cython_attribute("w", weights)

                # set delays if there were some
                self._set_delay(delays)

            # Other variables
            for var in desc['attributes']:
                if var == "w":
                    continue # already done

                try:
                    self._set_cython_attribute(var, desc[var])
                except Exception as e:
                    Messages._print(e)
                    Messages._warning('load(): the variable', var, 'does not exist in the current version of the network, skipping it.')
                    continue

            if connectivity_changed and not ConfigManager().get("suppress_warnings", self.net_id):
                Messages._info("Loading connectivity was successful, note that stored connectivity in save file diverges from the initial state ... (Projection{id} - {name})".format(id = self.id, name = self.name))

        # HD ( 5th Sep. 2024):
        #   This could path is only relevant for savefiles prior to version 4.8.0
        else:
            Messages._warning("load(): dendrites should be added in an ascending order for performance reasons.")
            Messages._warning("This might require some time to adapt the data structure ...")

            # (re-)initialize connectivity
            if isinstance(delays, (float, int)):
                delays = [[delays] for _ in range(len(desc['post_ranks']))] # wrapper expects list from list

            self.cyInstance.init_from_lil(desc['post_ranks'], desc['pre_ranks'], weights, delays, requires_sorting)

            for var in desc['attributes']:
                if var == "w":
                    continue # already done

                for lil_idx, post_rank in enumerate(desc['post_ranks']):
                    self.dendrite(post_rank).__setattr__(var, desc[var][lil_idx])

    ################################
    ## Structural plasticity
    ################################
    def start_pruning(self, period:float=None) -> None:
        """
        Starts pruning the synapses in the projection if the synapse defines a 'pruning' argument.

        'structural_plasticity' must be set to True in setup().

        :param period: how often pruning should be evaluated (default: dt, i.e. each step)
        """
        if not period:
            period = ConfigManager().get('dt', self.net_id)
        if not self.cyInstance:
            Messages._error('Can not start pruning if the network is not compiled.')

        if ConfigManager().get('structural_plasticity', self.net_id):
            try:
                self.cyInstance._pruning = True
                self.cyInstance._pruning_period = int(period/ConfigManager().get('dt', self.net_id))
                self.cyInstance._pruning_offset = Global.get_current_step()
            except :
                Messages._error("The synapse does not define a 'pruning' argument.")

        else:
            Messages._error("You must set 'structural_plasticity' to True in setup() to start pruning connections.")


    def stop_pruning(self) -> None:
        """
        Stops pruning the synapses in the projection if the synapse defines a 'pruning' argument.

        'structural_plasticity' must be set to True in `setup()`.
        """
        if not self.cyInstance:
            Messages._error('Can not stop pruning if the network is not compiled.')

        if ConfigManager().get('structural_plasticity', self.net_id):
            try:
                self.cyInstance._pruning = False
            except:
                Messages._error("The synapse does not define a 'pruning' argument.")

        else:
            Messages._error("You must set 'structural_plasticity' to True in setup() to start pruning connections.")

    def start_creating(self, period:float=None) -> None:
        """
        Starts creating the synapses in the projection if the synapse defines a 'creating' argument.

        'structural_plasticity' must be set to True in setup().

        :param period: how often creating should be evaluated (default: dt, i.e. each step)
        """
        if not period:
            period = ConfigManager().get('dt', self.net_id)
        if not self.cyInstance:
            Messages._error('Can not start creating if the network is not compiled.')

        if ConfigManager().get('structural_plasticity', self.net_id):
            try:
                self.cyInstance._creating = True
                self.cyInstance._creating_period = int(period/ConfigManager().get('dt', self.net_id))
                self.cyInstance._creating_offset = Global.get_current_step()
            except:
                Messages._error("The synapse does not define a 'creating' argument.")

        else:
            Messages._error("You must set 'structural_plasticity' to True in setup() to start creating connections.")

    def stop_creating(self) -> None:
        """
        Stops creating the synapses in the projection if the synapse defines a 'creating' argument.

        'structural_plasticity' must be set to True in setup().
        """
        if not self.cyInstance:
            Messages._error('Can not stop creating if the network is not compiled.')

        if ConfigManager().get('structural_plasticity', self.net_id):
            try:
                self.cyInstance._creating = False
            except:
                Messages._error("The synapse does not define a 'creating' argument.")

        else:
            Messages._error("You must set 'structural_plasticity' to True in setup() to start creating connections.")

    ################################
    # Paradigm specific functions
    ################################
    def update_launch_config(self, nb_blocks:int=-1, threads_per_block:int=32) -> None:
        """
        Allows the adjustment of the CUDA launch config (since 4.7.2).

        :param nb_blocks: number of CUDA blocks which can be 65535 at maximum. If set to -1, the number of launched blocks is computed by ANNarchy.
        :param threads_per_block: number of CUDA threads for one block which can be maximally 1024.
        """
        if not _check_paradigm("cuda", self.net_id):
            Messages._warning("Projection.update_launch_config() is intended for usage on CUDA devices")
            return

        if self.initialized:
            self.cyInstance.update_launch_config(nb_blocks=nb_blocks, threads_per_block=threads_per_block)
        else:
            Messages._error("Projection.update_launch_config() should be called after compile()")

    ################################
    ## Memory Management
    ################################
    def _size_in_bytes(self) -> int:
        """
        Returns the size in bytes of the allocated memory on C++ side. Note that this does not reflect monitored data and that it only works after compile() was invoked.
        """
        if self.initialized:
            return self.cyInstance.size_in_bytes()
        else:
            return 0

    def _clear(self):
        """
        Deallocates the container within the C++ instance. The population object is not usable anymore after calling this function.

        Warning: should be only called by the net deconstructor (in the context of parallel_run).
        """
        if self.initialized:
            self.cyInstance.clear()
            self.initialized = False



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

import os, sys, importlib
import subprocess
import shutil
import multiprocessing
import time
import json
import numpy as np

# ANNarchy core informations
import ANNarchy

from ANNarchy.intern.NetworkManager import NetworkManager
from ANNarchy.intern.Profiler import Profiler
from ANNarchy.intern.ConfigManagement import _update_global_config, _check_paradigm, ConfigManager
from ANNarchy.intern import Messages

from ANNarchy.extensions.bold.NormProjection import _update_num_aff_connections
from ANNarchy.generator.Template.CMakeTemplate import *
from ANNarchy.generator.CodeGenerator import CodeGenerator
from ANNarchy.generator.Sanity import check_structure, check_experimental_features
from ANNarchy.generator.Utils import check_cuda_version
from ANNarchy.parser.report.Report import report

from packaging.version import parse as parse_version

# String containing the extra libs which can be added by extensions
# e.g. extra_libs = ['-lopencv_core', '-lopencv_video']
extra_libs = []

def _folder_management(annarchy_dir, profile_enabled, clean, net_id):
    """
    ANNarchy is provided as a python package. For compilation a local folder
    'annarchy' is created in the current working directory.

    *Parameter*:

    * annarchy_dir : subdirectory
    * *profile_enabled*: copy needed data for profile extension
    """

    # Verbose
    if ConfigManager().get('verbose', net_id):
        Messages._print("Create subdirectory.")

    if clean or profile_enabled:
        shutil.rmtree(annarchy_dir, True)

    # Create the subdirectory
    if not os.path.exists(annarchy_dir):
        os.makedirs(annarchy_dir)
        os.mkdir(annarchy_dir+'/build')
        os.mkdir(annarchy_dir+'/generate')

    # Subdirectory for building networks
    if not os.path.exists(annarchy_dir+'/build/net'+str(net_id)):
        os.mkdir(annarchy_dir+'/build/net'+str(net_id))

    # Create the generate subfolder
    if not os.path.exists(annarchy_dir+'/generate/net'+str(net_id)):
        os.mkdir(annarchy_dir+'/generate/net'+str(net_id))

    # Save current ANNarchy version and paradigm
    with open(annarchy_dir+'/release', 'w') as wfile:
        wfile.write(ConfigManager().get('paradigm', net_id)+', '+ANNarchy.__release__)

    sys.path.append(annarchy_dir)

def compile(
        directory='annarchy',
        clean=False,
        compiler="default",
        compiler_flags="default",
        add_sources="",
        extra_libs="",
        cuda_config={'device': 0},
        annarchy_json="",
        silent=False,
        debug_build=False,
        trace_calls=None,
        profile_enabled=False,
        net_id=0
    ):
    """
    This method uses the network architecture to generate optimized C++ code and compile a shared library that will perform the simulation.

    The ``compiler``, ``compiler_flags`` and part of ``cuda_config`` take their default value from the configuration file ``~/.config/ANNarchy/annarchy.json``.

    The following arguments are for internal development use only:

    * **debug_build**: creates a debug version of ANNarchy, which logs the creation of objects and some other data (default: False).
    * **trace_calls**: if set to *init*, *simulate*, or *both* simulation calls inside of the C++ kernel are logged to console (default: None)
    * **profile_enabled**: creates a profilable version of ANNarchy, which logs several computation timings (default: False).

    :param directory: name of the subdirectory where the code will be generated and compiled. Default: "annarchy/".
    :param clean: boolean to specifying if the library should be recompiled entirely or only the changes since last compilation (default: False).
    :param compiler: C++ compiler to use. Default: g++ on GNU/Linux, clang++ on OS X. Valid compilers are [g++, clang++].
    :param compiler_flags: platform-specific flags to pass to the compiler. Defaults are defined in annarchy.json: "-march=native -O3". 
    :param cuda_config: dictionary defining the CUDA configuration for each population and projection.
    :param annarchy_json: compiler flags etc can be stored in a .json file normally placed in the home directory. With this flag one can directly assign a file location.
    :param silent: defines if status message like "Compiling... OK" should be printed.
    """
    # Check if the network has already been compiled
    if NetworkManager().get_network(net_id).compiled:
        Messages._print("compile(): the network has already been compiled, doing nothing.")
        return

    # Get command-line arguments. Note that setup() related flags has been partially parsed!
    options, unknown = ANNarchy._arg_parser.parser.parse_known_args()

    # Check for unknown flags
    if len(unknown) > 0 and ConfigManager().get('verbose', net_id):
        Messages._warning('unrecognized command-line arguments:', unknown)

    # Get CUDA configuration
    if options.gpu_device >= 0:
        cuda_config['device'] = int(options.gpu_device)

    # Check that a single backend is chosen
    if (options.num_threads != None) and (options.gpu_device >= 0):
        Messages._error('CUDA and openMP can not be active at the same time, please check your command line arguments.')

    # check if profiling was enabled by --profile
    if options.profile != None:
        profile_enabled = options.profile
        _update_global_config('profiling', options.profile)
        _update_global_config('profile_out', options.profile_out)

    # check if profiling enabled due compile()
    if profile_enabled != False and options.profile is None:
        _update_global_config('profiling', True)
    # if profiling is enabled
    if profile_enabled:
        # this will automatically create a globally available Profiler instance
        Profiler().enable_profiling()
        if ConfigManager().get('profile_out', net_id) is None:
            _update_global_config('profile_out', '.')

    # Debug the simulation kernel
    if debug_build is False:
        debug_build = options.debug  # debug build
    _update_global_config('debug', debug_build)

    # Trace function calls in simulation kernel
    if trace_calls is None:
        trace_calls = options.trace_calls  # debug build
    _update_global_config('trace_calls', trace_calls)

    # Clean
    clean = options.clean or clean # enforce rebuild

    # Compiling directory
    annarchy_dir = os.path.abspath(directory)
    if not annarchy_dir.endswith('/'):
        annarchy_dir += '/'

    # using a raw-string we can handle whitespaces in folder paths
    annarchy_dir = r'{}'.format(annarchy_dir)

    # Test if the current ANNarchy version is newer than what was used to create the subfolder
    if os.path.isfile(annarchy_dir+'/release'):
        with open(annarchy_dir+'/release', 'r') as rfile:
            prev_release = rfile.read().strip()
            prev_paradigm = ''

            # HD (03.08.2016):
            # in ANNarchy 4.5.7b I added also the paradigm to the release tag.
            # This if clause can be removed in later releases (TODO)
            if prev_release.find(',') != -1:
                prev_paradigm, prev_release = prev_release.split(', ')
            else:
                # old release tag
                clean = True

            if parse_version(prev_release) < parse_version(ANNarchy.__release__):
                clean = True

            elif prev_paradigm != ConfigManager().get('paradigm', net_id):
                clean = True

    else:
        clean = True # for very old versions

    # Check if the last compilation was successful
    if os.path.isfile(annarchy_dir+'/compilation'):
        with open(annarchy_dir + '/compilation', 'r') as rfile:
            res = rfile.read()
            if res.strip() == "0": # the last compilation failed
                clean = True
    else:
        clean = True

    # Manage the compilation subfolder
    _folder_management(annarchy_dir, profile_enabled, clean, net_id)

    # Create a Compiler object
    compiler = Compiler(
        annarchy_dir=annarchy_dir,
        clean=clean,
        compiler=compiler,
        compiler_flags=compiler_flags,
        add_sources=add_sources,
        extra_libs=extra_libs,
        path_to_json=annarchy_json,
        silent=silent,
        cuda_config=cuda_config,
        debug_build=debug_build,
        trace_calls=trace_calls,
        profile_enabled=profile_enabled,
        net_id=net_id
    )

    # Code Generation
    compiler.generate()

    if ConfigManager().get('verbose', net_id):
        net_str = "" if compiler.net_id == 0 else str(compiler.net_id)+" "
        Messages._print('Construct network '+net_str+'...', end=" ")

    # Create the Python objects
    _instantiate(compiler.net_id, cuda_config=compiler.cuda_config, user_config=compiler.user_config)

    # NormProjections require an update of afferent projections
    _update_num_aff_connections(compiler.net_id)

    if ConfigManager().get('verbose', net_id):
        Messages._print('OK')

    # Create a report if requested
    if options.report is not None:
        report(options.report)

def detect_cython():
    """
    Detect cython compiler and return absolute path.
    """
    # Check cython version
    with subprocess.Popen(sys.base_prefix + "/bin/cython%(major)s -V > /dev/null 2> /dev/null" % {'major': str(sys.version_info[0])}, shell=True) as test:
        if test.wait() != 0:
            cython = sys.base_prefix + "/bin/cython"
        else:
            cython = sys.base_prefix + "/bin/cython" + str(sys.version_info[0])
    # If not in the same folder as python, use the default
    with subprocess.Popen("%(cython)s -V > /dev/null 2> /dev/null" % {'cython': cython}, shell=True) as test:
        if test.wait() != 0:
            cython = shutil.which("cython"+str(sys.version_info[0]))
            if cython is None:
                cython = shutil.which("cython")
                if cython is None:
                    Messages._error("Unable to detect the path to cython.")

    return cython

def detect_cuda_arch():
    """
    For best performance, the compute compability should be mentioned to the compiler. CMake > 3.18 also enforces the
    setting of the compute-compability (see "cmake --help-policy CMP0104" for more details).
    """
    # I don't know ...
    if sys.platform.startswith('darwin'):
        return ""

    try:
        # check nvidia-smi for GPU details (only available for CUDA SDK > 11.6)
        query_result = subprocess.check_output("nvidia-smi --query-gpu=compute_cap --format=csv", shell=True)
    except:
        return ""

    # bytes to string conversion, the result contains compute_cap\nCC for each gpu\n
    query_result = query_result.decode('utf-8').split('\n')

    # NVIDIA and it's version numbering ...
    CC = int(float(query_result[1])*10)
    return """
    if(NOT DEFINED CMAKE_CUDA_ARCHITECTURES)
      set(CMAKE_CUDA_ARCHITECTURES {})
    endif()
""".format(CC)

class Compiler(object):
    " Main class to generate C++ code efficiently"

    def __init__(self, 
                 annarchy_dir, 
                 clean, 
                 compiler, 
                 compiler_flags, 
                 add_sources, 
                 extra_libs, 
                 path_to_json, 
                 silent, 
                 cuda_config, 
                 debug_build,
                 trace_calls,
                 profile_enabled, 
                 net_id):

        # Store arguments
        self.annarchy_dir = annarchy_dir
        self.clean = clean
        self.compiler = compiler
        self.compiler_flags = compiler_flags
        self.add_sources = add_sources
        self.extra_libs = extra_libs
        self.silent = silent
        self.cuda_config = cuda_config
        self.debug_build = debug_build
        self.trace_calls = trace_calls
        self.profile_enabled = profile_enabled
        self.net_id = net_id

        # Network to compile
        self.network = NetworkManager().get_network(net_id)

        # Aside from arguments provided to compile, some configuration is stored in annarchy.json
        if len(path_to_json) == 0:
            # check home-directory
            if os.path.exists(os.path.expanduser('~/.config/ANNarchy/annarchy.json')):
                with open(os.path.expanduser('~/.config/ANNarchy/annarchy.json'), 'r') as rfile:
                    self.user_config = json.load(rfile)
            else:
                # Set default user-defined config
                self.user_config = {
                    'openmp': {
                        'compiler': 'clang++' if sys.platform == "darwin" else 'g++',
                        'flags' : "-march=native -O3",
                    },
                    'cuda': {
                        'compiler': "nvcc",
                        'device': 0
                    }
                }

        else:
            # Load user-defined annarchy.json
            with open(path_to_json, 'r') as rfile:
                self.user_config = json.load(rfile)

        # Sanity check if the NVCC compiler is available
        if _check_paradigm("cuda", self.net_id):
            cmd = self.user_config['cuda']['compiler'] + " --version 1> /dev/null"

            if os.system(cmd) != 0:
                Messages._error("CUDA is not available on your system. Please check the CUDA installation or the annarchy.json configuration.")

            self.cuda_config['cuda_version'] = check_cuda_version(self.user_config['cuda']['compiler'])

    def generate(self):
        "Perform the code generation for the C++ code and create the Makefile."

        if Profiler().enabled or ConfigManager().get('show_time', self.net_id):
            t0 = time.time()
            if Profiler().enabled:
                Profiler().add_entry(t0, t0, "overall", "compile")

        if ConfigManager().get('verbose', self.net_id):
            net_str = "" if self.net_id == 0 else str(self.net_id)+" "
            Messages._print('Code generation '+net_str+'...', end=" ", flush=True)

        # Check that everything is allright in the structure of the network.
        check_structure(self.network.get_populations(), self.network.get_projections())

        # check if the user access some new features, or old ones which changed.
        check_experimental_features(self.network.get_populations(), self.network.get_projections())

        # Generate the code
        self.code_generation()

        # Generate the Makefile
        self.generate_makefile()

        # Copy the files if needed
        changed = self.copy_files()

        # Code generation done
        if ConfigManager().get('verbose', self.net_id):
            t1 = time.time()
            if not ConfigManager().get('show_time', self.net_id):
                Messages._print("OK", flush=True)
            else:
                Messages._print("OK (took "+str(t1-t0)+" seconds)", flush=True)

        # Shared libraries have os-dependent suffixes
        if sys.platform.startswith('linux'):
            lib_path = self.annarchy_dir + '/ANNarchyCore' + str(self.net_id) + '.so'
        elif sys.platform.startswith('darwin'):
            lib_path = self.annarchy_dir + '/ANNarchyCore' + str(self.net_id) + '.dylib'
        else:
            raise NotImplementedError

        # Perform compilation if something has changed
        if changed or not os.path.isfile(lib_path):
            self.compilation()

        # Set the compilation directory in the networks
        if ConfigManager().get('debug', self.net_id) or ConfigManager().get('disable_shared_library_time_offset', self.net_id):
            # In case of debugging or high-throughput simulations we want to
            # disable the trick below
            self.network.directory = self.annarchy_dir
        else:
            # Store the library in random subfolder
            # We circumvent with this an issue with reloading of shared libraries
            # see PEP 489: (https://www.python.org/dev/peps/pep-0489/) for more details
            directory = self.annarchy_dir+'/run_'+str(time.time())
            self.network.directory = directory
            os.mkdir(directory)
            shutil.copy(lib_path, directory)

        # Tell the networks they have been compiled
        self.network.compiled = True
        if Profiler().enabled:
            t1 = time.time()
            Profiler().update_entry(t0, t1, "overall", "compile")

    def copy_files(self):
        " Copy the generated files in the build/ folder if needed."
        changed = False
        if self.clean:
            for file in os.listdir(self.annarchy_dir+'/generate/net'+ str(self.net_id)):
                if file.endswith(".log"):
                    continue

                shutil.copy(self.annarchy_dir+'/generate/net'+ str(self.net_id) + '/' + file, # src
                            self.annarchy_dir+'/build/net'+ str(self.net_id) + '/' + file # dest
                           )
            changed = True

        else: # only the ones which have changed
            import filecmp
            for file in os.listdir(self.annarchy_dir+'/generate/net'+ str(self.net_id)):
                if file.endswith(".log"):
                    continue

                if not os.path.isfile(self.annarchy_dir+'/build/net'+ str(self.net_id) + '/' + file) or \
                    not file == "codegen.log" and \
                    not filecmp.cmp(self.annarchy_dir+'/generate/net' + str(self.net_id) + '/' + file,
                                    self.annarchy_dir+'/build/net'+ str(self.net_id) + '/' + file):


                    shutil.copy(self.annarchy_dir+'/generate//net'+ str(self.net_id) + '/' + file, # src
                                self.annarchy_dir+'/build/net'+ str(self.net_id) + '/' +file # dest
                               )
                    changed = True

                    if ConfigManager().get('verbose', self.net_id):
                        print(file, 'has changed')
                        # For debugging
                        # with open(self.annarchy_dir+'/generate/net'+ str(self.net_id) + '/' + file, 'r') as rfile:
                        #     text = rfile.read()
                        #     print(text)

            # Needs to check now if a file existed before in build/net but not in generate anymore
            for file in os.listdir(self.annarchy_dir+'/build/net'+ str(self.net_id)):
                if file == 'Makefile':
                    continue
                if file.endswith(".log"):
                    continue
                basename, extension = os.path.splitext(file)
                if not extension in ['h', 'hpp', 'cpp', 'cu']: # ex: .o
                    continue
                if not os.path.isfile(self.annarchy_dir+'/generate/net'+ str(self.net_id) + '/' + file):
                    if file.startswith('ANNarchyCore'):
                        continue
                    os.remove(self.annarchy_dir+'/build/net'+ str(self.net_id) + '/' + file)
                    if os.path.isfile(self.annarchy_dir+'/build/net'+ str(self.net_id) + '/' + basename + '.o'):
                        os.remove(self.annarchy_dir+'/build/net'+ str(self.net_id) + '/' + basename + '.o')
                    changed = True

        return changed

    def compilation(self):
        """ Create ANNarchyCore.so and py extensions if something has changed. """
        # STDOUT
        if not self.silent:
            if ConfigManager().get('verbose', self.net_id):
                msg = 'Compiling with ' + self.compiler + ' ' + self.compiler_flags
            else:
                msg = 'Compiling '
            if self.net_id > 0:
                msg += 'network ' + str(self.net_id)
            msg += '...'
            Messages._print(msg, end=" ", flush=True)
            if ConfigManager().get('show_time', self.net_id) or Profiler().enabled:
                t0 = time.time()

        target_dir = self.annarchy_dir + '/build/net'+ str(self.net_id)

        # using a raw-string we can handle whitespaces in folder paths
        target_dir = r'{}'.format(target_dir)

        # Switch to the build directory
        cwd = os.getcwd()
        os.chdir(target_dir)
        
        # CMake is quite talky by default (printing out compiler versions etc.)
        # We reduce the number of printed messages except the user enabled verbose mode.
        verbose = "> compile_stdout.log 2> compile_stderr.log" if not ConfigManager().get('verbose', self.net_id) else ""

        # Generate the Makefile from CMakeLists
        make_process = subprocess.Popen("cmake -S \"{}\" -B \"{}\" {}".format(target_dir, target_dir, verbose), shell=True)
        if make_process.wait() != 0:
            Messages._error('CMake generation failed.')


        # Start the compilation
        verbose = "> compile_stdout.log 2> compile_stderr.log" if not ConfigManager().get('verbose', self.net_id) else ""
        make_process = subprocess.Popen("make -j4" + verbose, shell=True)

        # Check for errors
        if make_process.wait() != 0:
            with open('compile_stderr.log', 'r') as rfile:
                msg = rfile.read()
            with open(self.annarchy_dir + '/compilation', 'w') as wfile:
                wfile.write("0")
            Messages._print(msg)
            try:
                if sys.platform.startswith('linux'):
                    os.remove('ANNarchyCore'+str(self.net_id)+'.so')
                elif sys.platform.startswith('darwin'):
                    os.remove('ANNarchyCore'+str(self.net_id)+'.dylib')
                else:
                    raise NotImplementedError
            except:
                pass
            Messages._error('Compilation failed.')
            
        else: # Note that the last compilation was successful
            with open(self.annarchy_dir + '/compilation', 'w') as wfile:
                wfile.write("1")

        # Return to the current directory
        os.chdir(cwd)

        if not self.silent:
            t1 = time.time()

            if not ConfigManager().get('show_time', self.net_id):
                Messages._print('OK', flush=True)
            else:
                Messages._print('OK (took '+str(t1 - t0)+'seconds.', flush=True)

            if Profiler().enabled:
                Profiler().add_entry(t0, t1, "compilation", "compile")

    def generate_makefile(self):
        """
        Generate the Makefile.

        The makefile consists of two stages compile the cython wrapper and
        compile the ANNarchy model files. Both is then linked together to
        a shared library usable in Python.
        """
        # Compiler
        if self.compiler == "default":
            self.compiler = self.user_config['openmp']['compiler']
        if self.compiler_flags == "default":
            self.compiler_flags = self.user_config['openmp']['flags']

        # flags are common to all platforms
        if not self.debug_build:
            cpu_flags = self.compiler_flags
        else:
            cpu_flags = "-O0 -g -D_DEBUG -march=native"

        # trace of function calls should be possible without debug mode
        if self.trace_calls is not None:
            if self.trace_calls in ["init", "both"]:
                cpu_flags += " -D_TRACE_INIT"

            if self.trace_calls in ["simulate", "both"]:
                cpu_flags += " -D_TRACE_SIMULATION_STEPS"

        if self.profile_enabled:
            cpu_flags += " -g"

        # OpenMP flag
        omp_flag = ""
        if ConfigManager().get('paradigm', self.net_id) == "openmp" :
            omp_flag = "-fopenmp"

        # Disable openMP parallel RNG?
        if ConfigManager().get('disable_parallel_rng', self.net_id) and _check_paradigm("openmp", self.net_id):
            cpu_flags += " -D_DISABLE_PARALLEL_RNG "

        # Disable auto-vectorization
        if ConfigManager().get('disable_SIMD_Eq', self.net_id) and _check_paradigm("openmp", self.net_id):
            cpu_flags += " -fno-tree-vectorize"

        # Cuda Library and Compiler
        #
        # hdin (22.03.2016): we should verify in the future, if compute_35 remains as best
        # configuration for Keplar and upwards.
        cuda_gen = ""
        gpu_flags = ""
        gpu_compiler = "nvcc"
        gpu_ldpath = ""
        xcompiler_flags = ""
        if sys.platform.startswith('linux') and ConfigManager().get('paradigm', self.net_id) == "cuda":
            cuda_gen = "" # TODO: -arch sm_%(ver)s

            if self.debug_build:
                gpu_flags = "-g -G -D_DEBUG"

            # read the config file for the cuda lib path
            if 'cuda' in self.user_config.keys():
                gpu_compiler = self.user_config['cuda']['compiler']
                gpu_ldpath = '-L' + self.user_config['cuda']['path'] + '/lib'
                gpu_flags += self.user_config['cuda']['flags']

            # -Xcompiler expects the arguments seperated by ','
            if len(cpu_flags.strip()) > 0:
                xcompiler_flags = cpu_flags.replace(" ",",")

        # Extra libs from extensions such as opencv
        libs = self.extra_libs
        for lib in extra_libs:
            libs += str(lib) + ' '


        if ConfigManager().get('paradigm', self.net_id) == "cuda":
            set_cuda_arch = detect_cuda_arch()
        else:
            set_cuda_arch = ""

        # Include path to Numpy is not standard on all distributions
        numpy_include = np.get_include()

        # ANNarchy default header: sparse matrix formats
        annarchy_include = ANNarchy.__path__[0]+'/include'

        # Thirdparty includes (C++ files)
        thirdparty_include = ANNarchy.__path__[0]+'/thirdparty'

        # The connector module needs to reload some header files,
        # ANNarchy.__path__ provides the installation directory
        path_to_cython_ext = '-I'+ANNarchy.__path__[0]+'/core/cython_ext/\" \"-I'+ANNarchy.__path__[0][:-8]

        # Create Makefiles depending on the target platform and parallel framework
        if sys.platform.startswith('linux'): # Linux systems
            if ConfigManager().get('paradigm', self.net_id) == "cuda":
                makefile_template = linux_cuda_template
            else:
                makefile_template = linux_omp_template

        elif sys.platform == "darwin":   # mac os
            if self.compiler == 'clang++':
                makefile_template = osx_clang_template
                if ConfigManager().get('num_threads', self.net_id) == 1: # clang should report that it does not support openmp
                    omp_flag = ""
            else:
                makefile_template = osx_gcc_template

        else: 
            # Windows: to test....
            Messages._warning("Compilation on windows is not supported yet. We recommend to use WSL on windows systems.")

        # Gather all Makefile flags
        makefile_flags = {
            'compiler': self.compiler,
            'add_sources': self.add_sources,
            'cpu_flags': cpu_flags,
            'cuda_gen': cuda_gen,
            'gpu_compiler': gpu_compiler,
            'gpu_flags': gpu_flags,
            'xcompiler_flags': xcompiler_flags,
            'gpu_ldpath': gpu_ldpath,
            'openmp': omp_flag,
            'set_cuda_arch': set_cuda_arch,
            'extra_libs': libs,
            'numpy_include': numpy_include,
            'annarchy_include': annarchy_include,
            'thirdparty_include': thirdparty_include,
            'net_id': self.net_id,
            'cython_ext': path_to_cython_ext
        }

        # Write the Makefile to the disk
        with open(self.annarchy_dir + '/generate/net'+ str(self.net_id) + '/CMakeLists.txt', 'w') as wfile:
            wfile.write(makefile_template % makefile_flags)


    def code_generation(self):
        """
        Code generation dependent on paradigm
        """
        # First, we remove previously created files
        target_folder = self.annarchy_dir+'/generate/net'+ str(self.net_id)
        if ConfigManager().get("verbose", self.net_id):
            print('\n\nCheck target folder:', target_folder)
        for file in os.listdir(target_folder):
            if ConfigManager().get("verbose", self.net_id):
                print("  - remove file:", file)
            os.remove(target_folder+'/'+file)

        # Then, we generate the code for the current network
        generator = CodeGenerator(
            self.annarchy_dir, 
            self.net_id, 
            self.cuda_config)
        
        generator.generate()


def load_cython_lib(libname, libpath):
    """
    Load the shared library created by Cython using importlib. Follows the example
    "Multiple modules in one library" in PEP 489.

    TODO:

    As described in PEP 489 "Module Reloading" a reloading of dynamic extension modules is
    not supported. This leads to some problems for our reusage of the ANNarchyCore library ...

    Sources:

    PEP 489: (https://www.python.org/dev/peps/pep-0489/)
    """
    # create a loader to mimic find module
    loader = importlib.machinery.ExtensionFileLoader(libname, libpath)
    spec = importlib.util.spec_from_loader(libname, loader)
    module = importlib.util.module_from_spec(spec)

    loader.exec_module(module)

    return module

def _instantiate(net_id, import_id=-1, cuda_config=None, user_config=None, core_list=None):
    """
    After every is compiled, actually create the Cython objects and bind them to the Python ones.
    """
    if Profiler().enabled:
        t0 = time.time()
        Profiler().add_entry(t0, t0, "overall", "instantiate") # placeholder, to have the correct ordering

    # parallel_run(number=x) defines multiple networks (net_id) but only network0 is compiled
    if import_id < 0:
        import_id = net_id

    # subdirectory where the library lies
    annarchy_dir = NetworkManager().get_network(net_id=import_id).directory
    libname = 'ANNarchyCore' + str(import_id)
    if sys.platform.startswith('linux'):
        libpath = annarchy_dir + '/' + libname + '.so'
    elif sys.platform.startswith('darwin'):
        libpath = annarchy_dir + '/' + libname + '.dylib'
    else:
        raise NotImplementedError

    # Load the library
    cython_module = load_cython_lib(libname, libpath)
    NetworkManager().get_network(net_id=net_id).instance = cython_module

    # Sets the desired number of threads and execute thread placement.
    # This must be done before any other objects are initialized.
    if _check_paradigm("openmp", net_id):
        # check for global setting
        if core_list is None:
            core_list = ConfigManager().get('visible_cores', net_id)

        # the user configured a setup
        if core_list != []:
            # some sanity check
            if len(core_list) > multiprocessing.cpu_count():
                Messages._error("The length of core ids provided to setup() is larger than available number of cores")

            if len(core_list) < ConfigManager().get('num_threads', net_id):
                Messages._error("The list of visible cores should be at least the number of cores.")

            if np.amax(np.array(core_list)) > multiprocessing.cpu_count():
                Messages._error("At least one of the core ids provided to setup() is larger than available number of cores")

            if len(core_list) != len(list(set(core_list))):
                Messages._warning("The provided core list contains doubled entries - is this intended?")

            cython_module.set_number_threads(ConfigManager().get('num_threads', net_id), np.array(core_list))

        else:
            # HD (26th Oct 2020): the current version of psutil only consider one CPU socket
            #                     but there is a discussion of adding multi-sockets, so we could
            #                     re-add this code later ...
            """
            num_cores = psutil.cpu_count(logical=False)
            # Check if the number of threads make sense
            if num_cores < ConfigManager().get('num_threads', net_id):
                Messages._warning("The number of threads =", ConfigManager().get('num_threads', net_id), "exceeds the number of available physical cores =", num_cores)

            # ANNarchy should run only on physical cpu cores
            core_list = np.arange(0, num_cores)
            """
            cython_module.set_number_threads(ConfigManager().get('num_threads', net_id), [])

        if ConfigManager().get('num_threads', net_id) > 1:
            if ConfigManager().get('verbose', net_id):
                Messages._print('Running simulation with', ConfigManager().get('num_threads', net_id), 'threads.')
        else:
            if ConfigManager().get('verbose', net_id):
                Messages._print('Running simulation single-threaded.')

    elif _check_paradigm("cuda", net_id):
        # check if there is a configuration,
        # otherwise fall back to default device
        device = 0
        if user_config is not None and 'cuda' in user_config.keys():
            device = int(cuda_config['device'])
        elif cuda_config is not None:
            device = int(user_config['cuda']['device'])

        if ConfigManager().get('verbose', net_id):
            Messages._print('Setting GPU device', device)

        # Set the CUDA device
        cython_module.set_device(device)

    else:
        raise NotImplementedError

    # Instantiate CPP objects
    cython_module.pyx_create()

    # Configure seeds for C++ random number generators
    # Required for state updates and also (in future) construction of connectivity
    seed = ConfigManager().get('seed', net_id)
    if seed is None:
        seed = int(time.time())

    if not ConfigManager().get('disable_parallel_rng', net_id):
        cython_module.set_seed(seed, ConfigManager().get('num_threads', net_id), ConfigManager().get('use_seed_seq', net_id))
    else:
        cython_module.set_seed(seed, 1, ConfigManager().get('use_seed_seq', net_id))

    if Profiler().enabled:
        # register the CPP profiling instance
        # Attention: since ANNarchy 5.0 this need to be instantiated before any other cpp object.
        Profiler()._cpp_profiler = NetworkManager().get_network(net_id).instance.Profiling_wrapper()

    # Bind the py extensions to the corresponding python objects
    for pop in NetworkManager().get_network(net_id).get_populations():
        if ConfigManager().get('verbose', net_id):
            Messages._print('Instantiate population ( name =', pop.name, ', size =', pop.size,')')
        if ConfigManager().get('show_time', net_id):
            t0 = time.time()

        # Instantiate the population
        pop._instantiate(cython_module)

        if ConfigManager().get('show_time', net_id):
            Messages._print('  instantiate of the popukatuib took', (time.time()-t0)*1000, 'milliseconds')

    # Instantiate projections
    for proj in NetworkManager().get_network(net_id).get_projections():
        if ConfigManager().get('verbose', net_id):
            Messages._print('Instantiate projection ( pre =', proj.pre.name, ', post =', proj.post.name, ', target =', proj.target, ')')
        if ConfigManager().get('show_time', net_id):
            t0 = time.time()

        # Create the projection
        proj._instantiate(cython_module)

        if ConfigManager().get('show_time', net_id):
            Messages._print('  instantiate of the projection took', (time.time()-t0)*1000, 'milliseconds')

    # Finish to initialize the network
    cython_module.pyx_initialize(ConfigManager().get('dt', net_id))

    # Set the user-defined constants
    for obj in NetworkManager().get_network(net_id).get_constants():
        getattr(cython_module, 'set_'+obj.name)(obj.value)

    # Transfer initial values
    for pop in NetworkManager().get_network(net_id).get_populations():
        if ConfigManager().get('verbose', net_id):
            Messages._print('Initializing C++ counterpart of population', pop.name)
        pop._init_attributes()
    for proj in NetworkManager().get_network(net_id).get_projections():
        if ConfigManager().get('verbose', net_id):
            Messages._print('Initializing C++ counterpart of projection', proj.name)
        proj._init_attributes()

    # Start the monitors
    for monitor in NetworkManager().get_network(net_id).get_monitors():
        monitor._init_monitoring()

    if Profiler().enabled:
        t1 = time.time()
        Profiler().update_entry(t0, t1, "overall", "instantiate")



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

from ANNarchy.intern.SpecificPopulation import SpecificPopulation
from ANNarchy.intern.ConfigManagement import ConfigManager
from ANNarchy.intern import Messages

from ANNarchy.core.Population import Population
from ANNarchy.core.Neuron import Neuron

class SpikeSourceArray(SpecificPopulation):
    """
    Spike source generating spikes at the times given in the `spike_times` array.

    Depending on the initial array provided, the population will have one or several neurons, but the geometry can only be one-dimensional.

    You can later modify the spike_times attribute of the population, but it must have the same number of neurons as the initial one.

    The spike times are by default relative to the start of a simulation.
    If you call the ``reset()`` method of a ``SpikeSourceArray``, this will set the spike times relative to the current time.
    You can then repeat a stimulation many times.

    ```python
    # 2 neurons firing at 100Hz with a 1 ms delay
    times = [
        [ 10, 20, 30, 40],
        [ 11, 21, 31, 41]
    ]
    inp = net.create(ann.SpikeSourceArray(spike_times=times))

    net.compile()

    # Spikes at 10/11, 20/21, etc
    net.simulate(50)

    # Reset the internal time of the SpikeSourceArray
    inp.reset()

    # Spikes at 60/61, 70/71, etc
    net.simulate(50)
    ```

    :param spike_times: a list of times at which a spike should be emitted if the population should have only 1 neuron, a list of lists otherwise. Times are defined in milliseconds, and will be rounded to the closest multiple of the discretization time step dt.
    :param name: optional name for the population.
    """
    def __init__(self, spike_times:list[float], name:str=None, copied=False, net_id=0):

        if not isinstance(spike_times, list):
            Messages._error('In a SpikeSourceArray, spike_times must be a Python list.')

        if isinstance(spike_times[0], list): # several neurons
            nb_neurons = len(spike_times)
        else: # a single Neuron
            nb_neurons = 1
            spike_times = [ spike_times ]

        # Create a fake neuron just to be sure the description has the correct parameters
        neuron = Neuron(
            parameters="",
            equations="",
            spike="t==0",
            reset="",
            name="Spike source",
            description="Spike source array."
        )

        SpecificPopulation.__init__(self, geometry=nb_neurons, neuron=neuron, name=name, copied=copied, net_id=net_id)

        self.init['spike_times'] = spike_times


    def _copy(self, net_id=None):
        "Returns a copy of the population when creating networks."
        return SpikeSourceArray(
            self.init['spike_times'], self.name, 
            copied=True, net_id=self.net_id if net_id is None else net_id)

    def _sort_spikes(self, spike_times):
        "Sort, unify the spikes and transform them into steps."
        return [
            sorted(
                list(
                    set(
                        [round(t/ConfigManager().get('dt', self.net_id)) for t in neur_times]
                    )
                )
            ) for neur_times in spike_times
        ]

    def _generate_st(self):
        """
        Code generation for single-thread.
        """
        self._generate_omp()

    def _generate_omp(self):
        """
        Code generation for openMP paradigm.
        """
        # Add possible targets
        for target in self.targets:
            tpl = {
                'name': 'g_%(target)s' % {'target': target},
                'locality': 'local',
                'eq': '',
                'bounds': {},
                'flags': [],
                'ctype': ConfigManager().get('precision', self.net_id),
                'init': 0.0,
                'transformed_eq': '',
                'pre_loop': {},
                'cpp': '',
                'switch': '',
                'untouched': {},
                'method': 'exponential',
                'dependencies': []
            }
            self.neuron_type.description['variables'].append(tpl)
            self.neuron_type.description['local'].append('g_'+target)

        self._specific_template['declare_additional'] = """
    // Custom local parameter spike_times
    // std::vector< %(float_prec)s > r ;
    std::vector< std::vector< long int > > spike_times ;
    std::vector< long int >  next_spike ;
    std::vector< int > idx_next_spike;
    long int _t;

    // Recompute the spike times
    void recompute_spike_times(){
        std::fill(next_spike.begin(), next_spike.end(), -10000);
        std::fill(idx_next_spike.begin(), idx_next_spike.end(), 0);
        for(int i=0; i< size; i++){
            if(!spike_times[i].empty()){
                int idx = 0;
                // Find the first spike time which is not in the past
                while(spike_times[i][idx] < _t){
                    idx++;
                }
                // Set the next spike
                if(idx < spike_times[i].size())
                    next_spike[i] = spike_times[i][idx];
                else
                    next_spike[i] = -10000;
            }
        }
    }
"""% { 'float_prec': ConfigManager().get('precision', self.net_id) }

        #self._specific_template['access_parameters_variables'] = ""

        self._specific_template['init_additional'] = """
        _t = 0;
        next_spike = std::vector<long int>(size, -10000);
        idx_next_spike = std::vector<int>(size, 0);
        this->recompute_spike_times();
""" 

        self._specific_template['reset_additional'] = """
        _t = 0;
        this->recompute_spike_times();
"""

        if ConfigManager().get('num_threads', self.net_id) == 1:
            self._specific_template['update_variables'] = """
        if(_active){
            spiked.clear();
            for(int i = 0; i < size; i++){
                // Emit spike
                if( _t == next_spike[i] ){
                    last_spike[i] = _t;
                    idx_next_spike[i]++ ;
                    if(idx_next_spike[i] < spike_times[i].size()){
                        next_spike[i] = spike_times[i][idx_next_spike[i]];
                    }
                    spiked.push_back(i);
                }
            }
            _t++;
        }
"""
        else:
            self._specific_template['update_variables'] = """
        if(_active){
            #pragma omp single
            {
                spiked.clear();
            }

            #pragma omp for
            for(int i = 0; i < size; i++){
                // Emit spike
                if( _t == next_spike[i] ){
                    last_spike[i] = _t;
                    idx_next_spike[i]++ ;
                    if(idx_next_spike[i] < spike_times[i].size()){
                        next_spike[i] = spike_times[i][idx_next_spike[i]];
                    }

                    #pragma omp critical
                    spiked.push_back(i);
                }
            }

            #pragma omp single
            {
                _t++;
            }
        }
"""
        self._specific_template['test_spike_cond'] = ""

        self._specific_template['wrapper'] = f"""
    // SpikeSourceArray pop{self.id}
    nanobind::class_<PopStruct{self.id}>(m, "pop{self.id}_wrapper")
        // Constructor
        .def(nanobind::init<int, int>())

        // Common attributes
        .def_rw("size", &PopStruct{self.id}::size)
        .def_rw("spike_times", &PopStruct{self.id}::spike_times)
        .def_rw("max_delay", &PopStruct{self.id}::max_delay)
        .def_rw("r", &PopStruct{self.id}::r)

        // Other methods
		.def("compute_firing_rate", &PopStruct{self.id}::compute_firing_rate)

        .def("activate", &PopStruct{self.id}::set_active)
        .def("reset", &PopStruct{self.id}::reset)
        .def("clear", &PopStruct{self.id}::clear);
        """

    def _generate_cuda(self):
        """
        Code generation for the CUDA paradigm.

        As the spike time generation is not a very compute intensive step but
        requires dynamic data structures, we don't implement it on the CUDA
        devices for now. Consequently, we use the CPU side implementation and
        transfer after computation the results to the GPU.
        """
        self._generate_st()

        # Maybe not the most elegant solution, but it works (HD: 14th Feb. 2025)
        part1 = self._specific_template['wrapper'].split("// Other methods")[0]
        part2 = self._specific_template['wrapper'].split("// Other methods")[1]
        self._specific_template['wrapper'] = part1
        self._specific_template['wrapper'] += f"""
        // Trigger host-to-device transfer
        .def_rw("r_host_to_device", &PopStruct{self.id}::r_host_to_device)

        // Other methods"""
        self._specific_template['wrapper'] += part2

        # attach transfer of spiked array to gpu
        # IMPORTANT: the outside transfer is necessary.
        # Otherwise, previous spike counts will be not reseted.
        self._specific_template['update_variables'] += """
        if ( _active ) {
            // Update Spike Count on GPU
            spike_count = spiked.size();
            cudaMemcpy( gpu_spike_count, &spike_count, sizeof(unsigned int), cudaMemcpyHostToDevice);

            // Transfer generated spikes to GPU
            if( spike_count > 0 ) {
                cudaMemcpy( gpu_spiked, spiked.data(), spike_count * sizeof(int), cudaMemcpyHostToDevice);
            }
        }
        """

        # overwrite default code generation for neural update
        self._specific_template['update_variable_body'] = ""
        self._specific_template['update_variable_invoke'] = ""
        self._specific_template['update_variable_header'] = ""
        self._specific_template['update_variable_call'] = """
    // host side update of neurons
    pop%(id)s->update();
""" % {'id': self.id}

        # overwrite default code generation for spike gather
        self._specific_template['spike_gather_body'] = ""
        self._specific_template['spike_gather_invoke'] = ""
        self._specific_template['spike_gather_header'] = ""
        self._specific_template['spike_gather_call'] = ""

    def _instantiate(self, module):
        # Create the Cython instance
        self.cyInstance = getattr(module, self.class_name+'_wrapper')(self.size, self.max_delay)
        self.cyInstance.spike_times = self._sort_spikes(self.init['spike_times'])

    def __setattr__(self, name, value):
        if name == 'spike_times':

            if not isinstance(value[0], list): # several neurons
                value = [ value ]
            
            if not len(value) == self.size:
                Messages._error('SpikeSourceArray: the size of the spike_times attribute must match the number of neurons in the population. Pad with `[]` if necessary.')

            self.init['spike_times'] = value # when reset is called
            if self.initialized:
                self.cyInstance.spike_times = self._sort_spikes(value)
        else:
            Population.__setattr__(self, name, value)

    def __getattr__(self, name):
        if name == 'spike_times':
            if self.initialized:
                return [ [ConfigManager().get('dt', self.net_id)*time for time in neur] for neur in self.cyInstance.spike_times]
            else:
                return self.init['spike_times']
        else:
            return Population.__getattribute__(self, name)


# SpecificPopulation inheritances
from .InputArray import InputArray
from .PoissonPopulation import PoissonPopulation
from .SpikeSourceArray import SpikeSourceArray
from .SpikeTrains import HomogeneousCorrelatedSpikeTrains
from .TimedArray import TimedArray
from .TimedPoissonPopulation import TimedPoissonPopulation

# SpecificProjecion inheritances
from .CurrentInjection import CurrentInjection
from .DecodingProjection import DecodingProjection

# For reporting
input_type_list = (
    InputArray,
    PoissonPopulation, 
    SpikeSourceArray, 
    TimedArray,
    HomogeneousCorrelatedSpikeTrains, 
    TimedPoissonPopulation,
    DecodingProjection, 
    CurrentInjection
)
input_name_list = [
    "Input Array", 
    "Poisson distribution", 
    "Spike source", 
    "Timed Array",
    "Homogeneous correlated spike trains",
    "Scheduled Poisson distribution", 
]

# Classes exported by the ANNarchy.inputs sub-module.
__all__ = [
    'InputArray',
    'PoissonPopulation', 
    'SpikeSourceArray', 
    'TimedArray',
    'HomogeneousCorrelatedSpikeTrains', 
    'TimedPoissonPopulation',
    'DecodingProjection', 
    'CurrentInjection'
]


"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

omp_population = {
    'template': """
class PopRecorder%(id)s : public Monitor
{
protected:
    int _id;

public:
    PopRecorder%(id)s(std::vector<int> ranks, int period, int period_offset, long int offset)
        : Monitor(ranks, period, period_offset, offset)
    {
    #ifdef _DEBUG
        std::cout << "PopRecorder%(id)s (this = " << this << ") instantiated." << std::endl;
    #endif
%(init_code)s

        // add monitor to global list
        this->_id = addRecorder(static_cast<Monitor*>(this));
    #ifdef _DEBUG
        std::cout << "PopRecorder%(id)s (" << this << ") received list position (ID) = " << this->_id << std::endl;
    #endif
    }

    ~PopRecorder%(id)s() {
    #ifdef _DEBUG
        std::cout << "PopRecorder%(id)s::~PopRecorder%(id)s(this = " << this << ")" << std::endl;
    #endif
    }

    void record() {
    #ifdef _TRACE_SIMULATION_STEPS
        std::cout << "PopRecorder%(id)s::record()" << std::endl;
    #endif
%(recording_code)s
    }

    void record_targets() {
%(recording_target_code)s
    }

    long int size_in_bytes() {
        long int size_in_bytes = 0;
%(size_in_bytes)s
        return size_in_bytes;
    }

%(clear_individual_container_code)s

    void clear() {
    #ifdef _DEBUG
        std::cout << "PopRecorder%(id)s::clear(this = " << this << ")" << std::endl;
    #endif
%(clear_all_container_code)s

        removeRecorder(this);
    }


%(struct_code)s
};
""",
    'local': {
    'struct': """
    // Local variable %(name)s
    std::vector< std::vector< %(type)s > > %(name)s ;
    bool record_%(name)s ; """,
    'init': """
        this->%(name)s = std::vector< std::vector< %(type)s > >();
        this->record_%(name)s = false; """,
    'recording': """
        if(this->record_%(name)s && ( (t - this->offset_) %% this->period_ == this->period_offset_ )){
        #ifdef _TRACE_SIMULATION_STEPS
            std::cout << "    Record '%(name)s' for pop%(id)s." << std::endl;
        #endif
            if(!this->partial) {
                this->%(name)s.push_back(pop%(id)s->%(name)s);
            } else {
                std::vector<%(type)s> tmp = std::vector<%(type)s>();
                for (unsigned int i=0; i<this->ranks.size(); i++){
                    tmp.push_back(pop%(id)s->%(name)s[this->ranks[i]]);
                }
                this->%(name)s.push_back(tmp);
            }
        }""",
    'size_in_bytes': """
// local variable %(name)s
size_in_bytes += sizeof(std::vector<%(type)s>) * %(name)s.capacity();
for(auto it=%(name)s.begin(); it!= %(name)s.end(); it++) {
    size_in_bytes += it->capacity() * sizeof(%(type)s);
}""",
    'clear': """
    void clear_%(name)s() {
        for(auto it = this->%(name)s.begin(); it != this->%(name)s.end(); it++) {
            it->clear();
            it->shrink_to_fit();
        }
        this->%(name)s.clear();
    }
    """
    },
    'semiglobal': { # Does not exist for populations
        'struct': "",
        'init': "",
        'recording': "",
        'size_in_bytes': "",
        'clear': ""
    },
    'global': {
        'struct': """
    // Global variable %(name)s
    std::vector< %(type)s > %(name)s ;
    bool record_%(name)s ; """,
        'init': """
        this->%(name)s = std::vector< %(type)s >();
        this->record_%(name)s = false; """,
        'recording': """
        if(this->record_%(name)s && ( (t - this->offset_) %% this->period_ == this->period_offset_ )){
            this->%(name)s.push_back(pop%(id)s->%(name)s);
        } """,
        'size_in_bytes': """
// global variable %(name)s
size_in_bytes += sizeof(%(type)s);""",
        'clear': """
    void clear_%(name)s() {
        this->%(name)s.clear();
    }
    """
    }
}

omp_projection = {
    'struct': """
class ProjRecorder%(id)s : public Monitor
{
protected:

    int _id;
    std::vector <int> indices;

public:

    ProjRecorder%(id)s(std::vector<int> ranks, int period, int period_offset, long int offset)
        : Monitor(ranks, period, period_offset, offset)
    {
    #ifdef _DEBUG
        std::cout << "ProjRecorder%(id)s (" << this << ") instantiated." << std::endl;
    #endif
        std::map< int, int > post_indices = std::map< int, int > ();
        auto post_rank = proj%(id)s->get_post_rank();

        for(int i=0; i<post_rank.size(); i++){
            post_indices[post_rank[i]] = i;
        }
        for(int i=0; i<this->ranks.size(); i++){
            this->indices.push_back(post_indices[this->ranks[i]]);
        }
        post_indices.clear();

        // initialize container
%(init_code)s

        // add monitor to global list
        this->_id = addRecorder(static_cast<Monitor*>(this));
    #ifdef _DEBUG
        std::cout << "ProjRecorder%(id)s (" << this << ") received list position (ID) = " << this->_id << std::endl;
    #endif

    };

    void record() {
%(recording_code)s
    };

    void record_targets() { /* nothing to do here */ }
    long int size_in_bytes() {
        size_t size_in_bytes = 0;

%(size_in_bytes_code)s

        return static_cast<long int>(size_in_bytes);
    }

    void clear() {
    #ifdef _DEBUG
        std::cout << "ProjRecorder%(id)s::clear(this = " << this << ")" << std::endl;
    #endif
%(clear_container_code)s
    }

%(clear_individual_code)s

%(struct_code)s
};
""",
    'local': {
        'struct': """
    // Local variable %(name)s
    std::vector< std::vector< std::vector< %(type)s > > > %(name)s ;
    bool record_%(name)s ;
""",
        'init' : """
        this->%(name)s = std::vector< std::vector< std::vector< %(type)s > > >();
        this->record_%(name)s = false;
""",
        'recording': """
        if(this->record_%(name)s && ( (t - this->offset_) %% this->period_ == this->period_offset_ )){
            std::vector< std::vector< %(type)s > > tmp;
            for(int i=0; i<this->ranks.size(); i++){
                tmp.push_back(std::move(proj%(id)s->get_matrix_variable_row<%(type)s>(proj%(id)s->%(name)s, this->indices[i])));
            }
            this->%(name)s.push_back(tmp);
            tmp.clear();
        }
""",
        'clear': """
    void clear_%(name)s() {
        for (auto it=%(name)s.begin(); it!= %(name)s.end(); it++) {
            for (auto it2=it->begin(); it2!= it->end(); it2++) {
                it2->clear();
                it2->shrink_to_fit();
            }
        }
        %(name)s.clear();
    }
""",
        'size_in_bytes': """
// local variable %(name)s
size_in_bytes += sizeof(std::vector<std::vector<%(type)s>>) * %(name)s.capacity();
for (auto it=%(name)s.begin(); it!= %(name)s.end(); it++) {
    for (auto it2=it->begin(); it2!= it->end(); it2++) {
        size_in_bytes += it2->capacity() * sizeof(%(type)s);
    }
}
"""
    },
    'semiglobal': {
        'struct': """
    // Semiglobal variable %(name)s
    std::vector< std::vector< %(type)s > > %(name)s ;
    bool record_%(name)s ;
""",
        'init' : """
        this->%(name)s = std::vector< std::vector< %(type)s > >();
        this->record_%(name)s = false;
""",
        'recording': """
        if(this->record_%(name)s && ( (t - this->offset_) %% this->period_ == this->period_offset_ )){
            std::vector< %(type)s > tmp;
            auto value = std::move(proj%(id)s->get_vector_variable_all<%(type)s>(proj%(id)s->%(name)s));
            for(int i=0; i<this->ranks.size(); i++){
                tmp.push_back(value[this->indices[i]]);
            }
            this->%(name)s.push_back(tmp);
            tmp.clear();
        }
""",
        'clear': """
    void clear_%(name)s() {
        for (auto it = %(name)s.begin(); it != %(name)s.end(); it++) {
            it->clear();
            it->shrink_to_fit();
        }
        %(name)s.clear();
        %(name)s.shrink_to_fit();
    }
""",
        'size_in_bytes': """
// semiglobal variable %(name)s
size_in_bytes += sizeof(std::vector<%(type)s>) * %(name)s.capacity();
for (auto it = %(name)s.begin(); it != %(name)s.end(); it++) {
    size_in_bytes += sizeof(%(type)s) * it->capacity();
}
"""
    },
    'global': {
        'struct': """
    // Global variable %(name)s
    std::vector< %(type)s > %(name)s ;
    bool record_%(name)s ;
""",
        'init' : """
        this->%(name)s = std::vector< %(type)s >();
        this->record_%(name)s = false;
""",
        'recording': """
        if(this->record_%(name)s && ( (t - this->offset_) %% this->period_ == this->period_offset_ )){
            this->%(name)s.push_back(proj%(id)s->%(name)s);
        }
""",
        'clear': """
    void clear_%(name)s() {
        %(name)s.clear();
        %(name)s.shrink_to_fit();
    }
""",
        'size_in_bytes': """
// global variable %(name)s
size_in_bytes += sizeof(%(type)s) * %(name)s.capacity();
"""
    }
}


"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

from ANNarchy.core import Global
from ANNarchy.intern.SpecificProjection import SpecificProjection
from ANNarchy.intern.ConfigManagement import ConfigManager
from ANNarchy.intern import Messages
from ANNarchy.models.Synapses import DefaultRateCodedSynapse, DefaultSpikingSynapse

class Transpose(SpecificProjection):
    """
    Transposed projection reusing the weights of an already-defined projection. 
    
    Even though the original projection can be learnable, this one can not. The computed post-synaptic potential is the default case for rate-coded projections: "w * pre.r"

    The proposed `target` can differ from the target of the forward projection.

    Example:

    ```python
    proj_ff = net.connect input, output, target="exc" )
    proj_ff.all_to_all(weights=Uniform(0,1)

    proj_fb = net.connect(Transpose(proj_ff, target="inh"))
    proj_fb.transpose()
    ```
    
    :param projection: original projection.
    :param target: type of the connection (can differ from the original one).
    """
    def __init__(self, projection, target, name=None, copied=False, net_id=0):

        # Transpose is not intended for hybrid projections
        if projection.pre.neuron_type.type == "rate" and projection.post.neuron_type.type == "rate":
            SpecificProjection.__init__(
                self,
                pre = projection.post,
                post = projection.pre,
                target = target,
                synapse = DefaultRateCodedSynapse,
                name = name,
                copied=copied,
                net_id=net_id,
            )
        elif projection.pre.neuron_type.type == "spike" and projection.post.neuron_type.type == "spike":
            SpecificProjection.__init__(
                self,
                pre = projection.post,
                post = projection.pre,
                target = target,
                synapse = DefaultSpikingSynapse,
                name = name,
                copied=copied,
                net_id=net_id,
            )
        else:
            Messages._error('TransposeProjection are not applyable on hybrid projections ...')

        # in the code generation we directly access properties of the
        # forward projection. Therefore we store the link here to have access in
        # self._generate()
        self.projection = projection

        if (projection._connection_delay > 0.0):
            Messages._error('TransposeProjection can not be applied on delayed projections yet ...')

        # simply copy from the forward view
        self.delays = projection._connection_delay
        self.max_delay = projection.max_delay
        self.uniform_delay = projection.uniform_delay

    def _copy(self, pre, post, net_id=None):
        "Returns a copy of the projection when creating networks. Internal use only."
        return Transpose(
            projection=self.projection, 
            target=self.target,
            copied=True,
            net_id = self.net_id if not net_id else net_id
        )

    def _create(self):
        pass

    def connect_transpose(self):
        return self.transpose()
    
    def transpose(self):
        # create fake LIL object to have the forward view in C++
        try:
            from ANNarchy.cython_ext.Connector import LILConnectivity
        except Exception as e:
            Messages._print(e)
            Messages._error('ANNarchy was not successfully installed.')

        lil = LILConnectivity(dt=ConfigManager().get('dt', self.net_id))
        lil.max_delay = self.max_delay
        lil.uniform_delay = self.uniform_delay
        self.connector_name = "Transpose"
        self.connector_description = "Transpose"

    def _connect(self, module):
        self.cyInstance = getattr(module, 'proj'+str(self.id)+'_wrapper')()
        return True

    def _generate(self):
        """
        Overrides default code generation. This function is called during the code generation procedure.
        """
        if self.synapse_type.type == "rate":
            self._generate_rate_coded()
        else:
            self._generate_spiking()

    def _generate_rate_coded(self):
        """
        Generates the transpose projection for rate-coded models.
        """
        #
        # C++ definition and PYX wrapper
        self._specific_template['struct_additional'] = """
extern ProjStruct%(fwd_id_proj)s* proj%(fwd_id_proj)s;    // Forward projection
""" % { 'fwd_id_proj': self.projection.id }

        self._specific_template['declare_connectivity_matrix'] = """
    // LIL connectivity (inverse of proj%(id)s)
    std::vector< int > inv_post_rank ;
    std::vector< std::vector< std::pair< int, int > > > inv_pre_rank ;
""" % {'float_prec': ConfigManager().get('precision', self.net_id), 'id': self.projection.id}
        self._specific_template['export_connector_call'] = ""

        # TODO: error message on setter?
        self._specific_template['access_connectivity_matrix'] = """
    // Accessor to connectivity data
    std::vector<int> get_post_rank() {
        return inv_post_rank;
    }
    size_t nb_synapses() {
        size_t size = 0;
        for (auto it = inv_pre_rank.cbegin(); it != inv_pre_rank.cend(); it++) {
            size += it->size();
        }
        return size;
    }
    int nb_dendrites() {
        return inv_post_rank.size();
    }
    int dendrite_size(int lil_idx) {
        return inv_pre_rank[lil_idx].size();
    }
"""
        self._specific_template['init_additional'] = """
        // Inverse connectivity to Proj%(fwd_id_proj)s
        auto inv_conn =  std::map< int, std::vector< std::pair<int, int> > > ();

        for (int i = 0; i < proj%(fwd_id_proj)s->pre_rank.size(); i++) {
            int post_rk = proj%(fwd_id_proj)s->post_rank[i];

            for (int j = 0; j < proj%(fwd_id_proj)s->pre_rank[i].size(); j++ ) {
                int pre_rk = proj%(fwd_id_proj)s->pre_rank[i][j];

                inv_conn[pre_rk].push_back(std::pair<int, int>(i, j));
            }
        }

        // keys are automatically sorted
        for (auto it = inv_conn.begin(); it != inv_conn.end(); it++ ) {
            inv_post_rank.push_back(it->first);
            inv_pre_rank.push_back(it->second);
        }

        // sanity check
        size_t inv_size = 0;
        for (int i = 0; i < inv_post_rank.size(); i++) {
            inv_size += inv_pre_rank[i].size();
        }

        assert( (proj%(fwd_id_proj)s->nb_synapses() == inv_size) );
""" % { 'fwd_id_proj': self.projection.id }


        # No attributes
        self._specific_template['declare_parameters_variables'] = ""
        self._specific_template['export_parameters_variables'] = ""
        self._specific_template['access_parameters_variables'] = ""

        # Override the monitor to avoid recording the weights
        self._specific_template['monitor_class'] = ""
        self._specific_template['monitor_export'] = ""
        self._specific_template['monitor_wrapper'] = ""

        self._specific_template['size_in_bytes'] = ""
        self._specific_template['clear_container'] = ""

        self._specific_template['wrapper'] = """
    // Transpose ProjStruct%(id_proj)s
    nanobind::class_<ProjStruct%(id_proj)s>(m, "proj%(id_proj)s_wrapper")
        // Constructor
        .def(nanobind::init<>())

        // Flags
        .def_rw("_transmission", &ProjStruct%(id_proj)s::_transmission)
        .def_rw("_axon_transmission", &ProjStruct%(id_proj)s::_axon_transmission)
        .def_rw("_update", &ProjStruct%(id_proj)s::_update)
        .def_rw("_plasticity", &ProjStruct%(id_proj)s::_plasticity)

        // Other methods
        .def("clear", &ProjStruct%(id_proj)s::clear);    
    
    """ %  {
                'id_proj': self.id,
                'id_copy': self.projection.id,
                'float_prec': ConfigManager().get('precision', self.net_id)
            }

        # The weight index depends on the
        # weight of the forward projection
        if self.projection._has_single_weight():
            weight_index = ""
        else:
            weight_index = "[post_idx][pre_idx]"

        #
        # PSP code
        self._specific_template['psp_code'] = """
        if (pop%(id_post)s->_active && _transmission) {
            %(omp_code)s
            for (int i = 0; i < inv_post_rank.size(); i++) {
                %(float_prec)s sum = 0.0;

                for (auto it = inv_pre_rank[i].begin(); it != inv_pre_rank[i].end(); it++) {
                    auto post_idx = it->first;
                    auto pre_idx = it->second;

                    sum += pop%(id_pre)s->r[proj%(fwd_id_proj)s->post_rank[post_idx]] * proj%(fwd_id_proj)s->w%(index)s;
                }
                pop%(id_post)s->_sum_%(target)s[inv_post_rank[i]] += sum;
            }
        }
""" % { 'float_prec': ConfigManager().get('precision', self.net_id),
        'target': self.target,
        'id_pre': self.pre.id,
        'id_post': self.post.id,
        'fwd_id_proj': self.projection.id,
        'index': weight_index,
        'omp_code': "" if ConfigManager().get('num_threads', self.net_id) == 1 else "#pragma omp for"
}

    def _generate_spiking(self):
        """
        Generates the transpose projection for spiking models.

        TODO: openMP
        """
        if ConfigManager().get('num_threads', self.net_id) > 1:
            Messages._error('TransposeProjection for spiking projections is only available for single-thread yet ...')

        # Which projection is transposed
        self._specific_template['struct_additional'] = """
extern ProjStruct%(fwd_id_proj)s *proj%(fwd_id_proj)s;    // Forward projection
""" % { 'fwd_id_proj': self.projection.id }

        # Connectivity
        self._specific_template['declare_connectivity_matrix'] = "" # reuse fwd proj data
        self._specific_template['access_connectivity_matrix'] = """
    std::vector<int> get_post_rank() {
        return proj%(fwd_id_proj)s->inv_post_rank;
    }
    size_t nb_synapses() {
        size_t size = 0;
        for (auto it = proj%(fwd_id_proj)s->inv_pre_rank.cbegin(); it != proj%(fwd_id_proj)s->inv_pre_rank.cend(); it++) {
            size += (it->second).size();
        }
        return size;
    }
    int nb_dendrites() {
        return proj%(fwd_id_proj)s->inv_post_rank.size();
    }
    int dendrite_size(int lil_idx) {
        int post_rank = proj%(fwd_id_proj)s->inv_post_rank[lil_idx];
        return proj%(fwd_id_proj)s->inv_pre_rank[post_rank].size();
    }
""" % { 'fwd_id_proj': self.projection.id }

        # The weight index depends on the
        # weight of the forward projection
        if self.projection._has_single_weight():
            weight_index = ""
        else:
            weight_index = "[post_idx][syn_idx]"

        # Computation
        self._specific_template['psp_prefix'] = ""
        self._specific_template['psp_code'] = """
        for (auto it = pop%(id_pre)s->spiked.cbegin(); it != pop%(id_pre)s->spiked.cend(); it++) {
            auto pos_it = std::find(proj%(fwd_id_proj)s->post_rank.cbegin(), proj%(fwd_id_proj)s->post_rank.cend(), *it);
            if (pos_it == proj%(fwd_id_proj)s->post_rank.end())
                continue;

            auto post_idx = std::distance(proj%(fwd_id_proj)s->post_rank.cbegin(), pos_it);

            for (int syn_idx = 0; syn_idx < proj%(fwd_id_proj)s->pre_rank[post_idx].size(); syn_idx++) {
                auto pre_idx = proj%(fwd_id_proj)s->pre_rank[post_idx][syn_idx];
                pop%(id_post)s->g_%(target)s[pre_idx] += proj%(fwd_id_proj)s->w%(weight_index)s;
            }
        }
""" % {
    'id_pre': self.pre.id,
    'id_post': self.post.id,
    'target': self.target,
    'fwd_id_proj': self.projection.id,
    'weight_index': weight_index
}
        
        self._specific_template['wrapper'] = """
    // Transpose ProjStruct%(id_proj)s
    nanobind::class_<ProjStruct%(id_proj)s>(m, "proj%(id_proj)s_wrapper")
        // Constructor
        .def(nanobind::init<>())

        // Flags
        .def_rw("_transmission", &ProjStruct%(id_proj)s::_transmission)
        .def_rw("_axon_transmission", &ProjStruct%(id_proj)s::_axon_transmission)
        .def_rw("_update", &ProjStruct%(id_proj)s::_update)
        .def_rw("_plasticity", &ProjStruct%(id_proj)s::_plasticity)

        // Other methods
        .def("clear", &ProjStruct%(id_proj)s::clear);    
    
    """ %  {
                'id_proj': self.id,
                'id_copy': self.projection.id,
                'float_prec': ConfigManager().get('precision', self.net_id)
            }
        
        # No attributes
        self._specific_template['declare_parameters_variables'] = ""
        self._specific_template['export_parameters_variables'] = ""
        self._specific_template['access_parameters_variables'] = ""

        # Override the monitor to avoid recording the weights
        self._specific_template['monitor_class'] = ""
        self._specific_template['monitor_export'] = ""
        self._specific_template['monitor_wrapper'] = ""

        self._specific_template['size_in_bytes'] = ""
        self._specific_template['clear_container'] = ""

    ##############################
    ## Override useless methods
    ##############################
    def _data(self):
        "Disable saving."
        desc = {}
        desc['post_ranks'] = self.post_ranks
        desc['attributes'] = self.attributes
        desc['parameters'] = self.parameters
        desc['variables'] = self.variables

        desc['dendrites'] = []
        desc['number_of_synapses'] = 0
        return desc

    def save_connectivity(self, filename):
        "Not available."
        Messages._warning('Transposed projections can not be saved.')
    def save(self, filename):
        "Not available."
        Messages._warning('Transposed projections can not be saved.')
    def load(self, filename):
        "Not available."
        Messages._warning('Transposed projections can not be loaded.')

    # TODO: maybe this functions would be helpful for debugging. Even though
    #       they will be time consuming as the matrix need to be constructed.
    #       (HD, 9th July 2020)
    def receptive_fields(self, variable = 'w', in_post_geometry = True):
        "Not available."
        Messages._warning('Transposed projections can not display receptive fields.')
    def connectivity_matrix(self, fill=0.0):
        "Not available."
        Messages._warning('Transposed projections can not display connectivity matrices.')



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

from ANNarchy.core.Population import Population
from ANNarchy.core.Monitor import Monitor
from ANNarchy.core import Global
from ANNarchy.intern.NetworkManager import NetworkManager
from ANNarchy.intern.ConfigManagement import ConfigManager
from ANNarchy.intern import Messages

from .BoldModel import BoldModel
from .PredefinedModels import balloon_RN
from .AccProjection import AccProjection

import inspect

class BoldMonitor(object):
    """
    Monitors the BOLD signal for several populations using a computational model.

    Returned by `Network.boldmonitor()`.

    The monitor can be started and stopped with `start()` and `stop()`. The recorded data is retrieved with `get()`.
    """
    def __init__(self,
            populations: list=None,
            bold_model: BoldModel = None,
            mapping: dict={'I_CBF': 'r'},
            scale_factor: list[float]=None,
            normalize_input: list[int]=None,
            recorded_variables: list[str]=None,
            start:bool=False,
            copied:bool=False,
            net_id:int=0,
        ):
        
        self.net_id = net_id

        if bold_model is None:
            bold_model = bold.balloon_RN

        # instantiate if necessary, please note that population will make a deepcopy on this objects
        if inspect.isclass(bold_model):
            bold_model = bold_model()

        # for reporting
        bold_model._model_instantiated = True

        # The usage of [] as default arguments in the __init__ call lead to strange side effects.
        # We decided therefore to use None as default and create the lists locally.
        if populations is None:
            Messages._error("Either a population or a list of populations must be provided to the BOLD monitor (populations=...)")
        if scale_factor is None:
            scale_factor = []
        if normalize_input is None:
            normalize_input = []
        if recorded_variables is None:
            recorded_variables = []

        # argument check
        if not(isinstance(populations, list)):
            populations = [populations]
        if not(isinstance(scale_factor, list)):
            scale_factor = [scale_factor]*len(populations)
        if not(isinstance(normalize_input, list)):
            normalize_input = [normalize_input]*len(populations)
        if isinstance(recorded_variables, str):
            recorded_variables = [recorded_variables]

        if len(scale_factor) > 0:
            if len(populations) != len(scale_factor):
                Messages._error("BoldMonitor: Length of scale_factor must be equal to number of populations")

        if len(normalize_input) > 0:
            if len(populations) != len(normalize_input):
                Messages._error("BoldMonitor: Length of normalize_input must be equal to number of populations")

        # Check mapping
        for target, input_var in mapping.items():
            if not target in bold_model._inputs:
                Messages._error("BoldMonitor: the key " + target + " of mapping is not part of the BOLD model.")

        # Check recorded variables
        if len(recorded_variables) == 0:
            recorded_variables = bold_model._output
        else:
            # Add the output variables (and remove doublons)
            l1 = bold_model._output
            l2 = [recorded_variables] if isinstance(recorded_variables, str) else recorded_variables
            recorded_variables = list(set(l2+l1))
            recorded_variables.sort()


        # Get the corresponding network
        self._net = NetworkManager().get_network(net_id=net_id)

        # Add the container to the object management
        self.id = self._net._add_extension(extension=self)


        if not copied:

            # create the population
            self._bold_pop = self._net.create(1, neuron=bold_model, name=bold_model.name )
            self._bold_pop.enabled = start

            # create the monitor
            self._monitor = self._net.monitor(self._bold_pop, recorded_variables, start=start)

            # create the projection(s)
            self._acc_proj = []

            if len(scale_factor) == 0:
                pop_overall_size = 0
                for _, pop in enumerate(populations):
                    pop_overall_size += pop.size

                # the conductance is normalized between [0 .. 1]. This scale factor
                # should balance different population sizes
                for _, pop in enumerate(populations):
                    scale_factor_conductance = float(pop.size)/float(pop_overall_size)
                    scale_factor.append(scale_factor_conductance)

            if len(normalize_input) == 0:
                normalize_input = [0] * len(populations)
                # TODO: can we check if users used NormProjections? If not, this will crash ...

            for target, input_var in mapping.items():
                for pop, scale, normalize in zip(populations, scale_factor, normalize_input):
                    Messages._debug("Creating ACCProjection between", pop.name, self._bold_pop.name)
                    tmp_proj = AccProjection(
                        pre = pop, 
                        post=self._bold_pop, 
                        target=target, 
                        variable=input_var, 
                        scale_factor=scale, 
                        normalize_input=normalize,
                        net_id = net_id
                    )
                    tmp_proj.all_to_all(weights= 1.0)

                    self._acc_proj.append(tmp_proj)

        else: # TODO check

            # instances are assigned by the copying instance
            self._bold_pop = None
            self._monitor = None
            self._acc_proj = []

        self.name = "bold_monitor"

        # store arguments for copy
        self._populations = populations
        self._bold_model = bold_model
        self._mapping = mapping
        self._scale_factor = scale_factor
        self._normalize_input = normalize_input
        self._recorded_variables = recorded_variables
        self._start = start

        # Finalize initialization
        self._initialized = True if not copied else False

    #
    #   MONITOR functions
    #
    def start(self):
        """
        Starts recording as in `ANNarchy.core.Monitor.start()`.
        """
        self._monitor.start()

        # enable ODEs
        self._bold_pop.cyInstance.activate(True)

        # check if we have projections with baseline
        for proj in self._acc_proj:
            if proj._normalize_input > 0:
                proj.cyInstance.start(int(proj._normalize_input/ConfigManager().get('dt', self.net_id)))

    def stop(self):
        """
        Stops recording as in `ANNarchy.core.Monitor.stop()`.
        """
        self._monitor.stop()

        # enable ODEs
        self._bold_pop.cyInstance.activate(False)

    def get(self, variable):
        """
        Retrieves recordings as in `ANNarchy.core.Monitor.get()`.
        """
        return self._monitor.get(variable)


    #
    #   POPULATION functions i. e. access to model parameter
    #

    # Method called when accessing an attribute.
    # We overload the default to allow access to monitor variables.
    def __getattr__(self, name):

        if name == '_initialized' or not hasattr(self, '_initialized'): # Before the end of the constructor
            return object.__getattribute__(self, name)

        if self._initialized:
            if self._bold_pop.initialized == False:
                Messages._error("BoldMonitor: attributes can not modified before compile()")

            if name in self._bold_pop.attributes:
                return getattr(self._bold_pop, name)

        return object.__getattribute__(self, name)

    # Method called when accessing an attribute.
    # We overload the default to allow access to monitor variables.
    def __setattr__(self, name, value):

        if name == '_initialized' or not hasattr(self, '_initialized'): # Before the end of the constructor
            return object.__setattr__(self, name, value)

        if self._initialized:
            if self._bold_pop.initialized == False:
                Messages._error("BoldMonitor: attributes can not modified before compile()")

            if name in self._bold_pop.attributes:
                setattr(self._bold_pop, name, value)
            else:
                raise AttributeError("the variable '"+str(name)+ "' is not an attribute of the bold model.")

        else:
            object.__setattr__(self, name, value)

    #
    # Destruction
    def _clear(self):
        pass
